# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for server task input fields."""

from types import SimpleNamespace
from typing import ClassVar, override

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import LookupChildType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.models.lookups import LookupResultArtifact
from debusine.db.tasks.tests.test_inputs import InputTests
from debusine.server.tasks.inputs import CopyCollectionItemsCopiesInput
from debusine.server.tasks.models import CopyCollectionItemsCopies
from debusine.tasks import TaskConfigError


class CopyCollectionItemsCopiesInputTests(
    InputTests[CopyCollectionItemsCopiesInput]
):
    """Tests for :py:class:`CopyCollectionItemsCopiesInput`."""

    source_workspace: ClassVar[Workspace]
    source_build_logs: ClassVar[Collection]
    target_build_logs: ClassVar[Collection]
    work_requests: ClassVar[list[WorkRequest]]
    build_log_artifacts: ClassVar[list[Artifact]]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_workspace = cls.playground.create_workspace(
            name="source", public=True
        )
        cls.source_build_logs = cls.playground.create_singleton_collection(
            CollectionCategory.PACKAGE_BUILD_LOGS,
            workspace=cls.source_workspace,
        )
        cls.target_build_logs = cls.playground.create_singleton_collection(
            CollectionCategory.PACKAGE_BUILD_LOGS,
        )
        cls.work_requests = [
            cls.playground.create_worker_task(workspace=cls.source_workspace)
            for _ in range(2)
        ]
        cls.build_log_artifacts = []
        for work_request, version in zip(cls.work_requests, ("1.0-1", "1.0-2")):
            build_log = cls.playground.create_build_log_artifact(
                version=version,
                workspace=cls.source_workspace,
                work_request=work_request,
            )
            cls.source_build_logs.manager.add_artifact(
                build_log,
                user=cls.playground.get_default_user(),
                variables={
                    "work_request_id": work_request.id,
                    "vendor": "debian",
                    "codename": "sid",
                    "architecture": "amd64",
                },
            )
            cls.build_log_artifacts.append(build_log)

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_none(self) -> None:
        self.task.data = SimpleNamespace(field=None)
        with self.assertRaises(AssertionError):
            self.resolve()

    def test_empty_list(self) -> None:
        self.task.data = SimpleNamespace(field=[])
        self.assertEqual(self.resolve(), [])

    def test_valid(self) -> None:
        self.task.data = SimpleNamespace(
            field=[
                CopyCollectionItemsCopies(
                    source_items={
                        "collection": self.source_build_logs.pk,
                        "category": ArtifactCategory.PACKAGE_BUILD_LOG,
                    },
                    target_collection=self.target_build_logs.pk,
                )
            ]
        )
        info = self.resolve()
        self.assertEqual(len(info), 1)
        self.assertEqual(info[0].label, "field[0]")
        self.assertEqual(
            len(info[0].source_items), len(self.build_log_artifacts)
        )
        for source_item, build_log_artifact in zip(
            info[0].source_items, self.build_log_artifacts
        ):
            self.assertIsDBInputLookupResult(
                source_item,
                LookupResultArtifact(
                    lookup=self.task.data.field[0].source_items.root[0],
                    parent_collection_lookup=(
                        f"{self.source_build_logs.pk}@collections"
                    ),
                    collection_item=self.source_build_logs.child_items.get(
                        artifact=build_log_artifact
                    ),
                    artifact=build_log_artifact,
                ),
                label="field[0].source_items",
            )
        self.assertIsDBInputCollectionSingle(
            info[0].target_collection,
            self.target_build_logs,
            lookup=self.task.data.field[0].target_collection,
            label="field[0].target_collection",
        )
        self.assertFalse(info[0].unembargo)
        self.assertFalse(info[0].replace)
        self.assertIsNone(info[0].name_template)
        self.assertIsNone(info[0].variables)

    def test_source_items_lookup_fails(self) -> None:
        nonexistent_lookup = f"nonexistent@{CollectionCategory.TEST}"
        self.task.data = SimpleNamespace(
            field=[
                CopyCollectionItemsCopies(
                    source_items={"collection": nonexistent_lookup},
                    target_collection=self.target_build_logs.pk,
                )
            ]
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field\[0\].source_items' in"
            rf" {self.task.workspace} failed: {nonexistent_lookup!r} does not"
            rf" exist or is hidden",
        ):
            self.resolve()

    def test_source_items_contain_collection(self) -> None:
        sub_collection = self.playground.create_collection(
            name="test", category=CollectionCategory.TEST
        )
        CollectionItem.objects.create_from_collection(
            sub_collection,
            parent_collection=self.source_build_logs,
            name="test",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        self.task.data = SimpleNamespace(
            field=[
                CopyCollectionItemsCopies(
                    source_items={
                        "collection": self.source_build_logs.pk,
                        "child_type": LookupChildType.COLLECTION,
                    },
                    target_collection=self.target_build_logs.pk,
                )
            ]
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Bad lookup result from field 'field\[0\].source_items' in"
            rf" {self.task.workspace}: Cannot copy entire collections",
        ):
            self.resolve()

    def test_target_collection_lookup_fails(self) -> None:
        nonexistent_lookup = f"nonexistent@{CollectionCategory.TEST}"
        self.task.data = SimpleNamespace(
            field=[
                CopyCollectionItemsCopies(
                    source_items={"collection": self.source_build_logs.pk},
                    target_collection=nonexistent_lookup,
                )
            ]
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field\[0\].target_collection' in"
            rf" {self.task.workspace} failed: {nonexistent_lookup!r} does not"
            rf" exist or is hidden",
        ):
            self.resolve()

    def test_get_artifact_ids(self) -> None:
        self.task.data = SimpleNamespace(
            field=[
                CopyCollectionItemsCopies(
                    source_items={"collection": self.source_build_logs.pk},
                    target_collection=self.target_build_logs.pk,
                )
            ]
        )
        self.assertEqual(
            self.field.get_artifact_ids(self.resolve()),
            [artifact.id for artifact in self.build_log_artifacts],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the ContextMiddleware."""

from collections.abc import Mapping
from typing import Any, TYPE_CHECKING, cast
from unittest import IsolatedAsyncioTestCase

from asgiref.typing import (
    ASGIReceiveCallable,
    ASGISendCallable,
    WebSocketReceiveEvent,
)
from django.http import HttpRequest, HttpResponse

from debusine.db.context import context
from debusine.server.middlewares.context import (
    ContextMiddleware,
    ContextMiddlewareChannels,
)
from debusine.test.django import TestCase

if TYPE_CHECKING:
    from channels.consumer import _ChannelScope

    from debusine.db.models import Scope


class ContextMiddlewareTests(TestCase):
    """Test ContextMiddleware."""

    def test_preserve_context(self) -> None:
        """Make sure context works as intended on sync views."""
        mock_scope = cast("Scope", 42)
        mock_scope_old = cast("Scope", 41)

        def view(request: HttpRequest) -> HttpResponse:  # noqa: ARG001, U100
            # Application context is reset at the beginning of the view
            self.assertIsNone(context.scope)

            # It can be set inside the view
            context._scope.set(mock_scope)
            self.assertEqual(context.scope, mock_scope)

            return HttpResponse()

        with context.local():
            # Simulate a scope left around by a previous request
            context._scope.set(mock_scope_old)

            wrapped_view = ContextMiddleware(view)
            wrapped_view(cast(HttpRequest, None))

            # Application context is left set
            self.assertEqual(context.scope, mock_scope)

        self.assertIsNone(context.scope)


class ContextMiddlewareChannelsTests(IsolatedAsyncioTestCase, TestCase):
    """Test ContextMiddlewareChannels."""

    @classmethod
    async def _noop_receive(cls) -> WebSocketReceiveEvent:  # pragma: no cover
        """No-op receive callback for testing."""
        return WebSocketReceiveEvent(
            type="websocket.receive", bytes=None, text=""
        )

    @classmethod
    async def _noop_send(cls, message: Mapping[str, Any]) -> None:
        """No-op send callback for testing."""

    async def test_preserve_context(self) -> None:
        """Make sure context works as intended on async views."""
        mock_scope = cast("Scope", 42)
        mock_scope_old = cast("Scope", 41)

        async def view(
            scope: "_ChannelScope",  # noqa: ARG001, U100
            receive: ASGIReceiveCallable,  # noqa: ARG001, U100
            send: ASGISendCallable,  # noqa: ARG001, U100
        ) -> None:
            # Application context is reset at the beginning of the view
            self.assertIsNone(context.scope)

            # It can be set inside the view
            context._scope.set(mock_scope)
            self.assertEqual(context.scope, mock_scope)

        with context.local():
            # Simulate a scope left around by a previous request
            context._scope.set(mock_scope_old)

            wrapped_view = ContextMiddlewareChannels(view)
            await wrapped_view(
                cast("_ChannelScope", None), self._noop_receive, self._noop_send
            )

            # Application context is left set
            self.assertEqual(context.scope, mock_scope)

        self.assertIsNone(context.scope)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used for asset data."""

from debusine.assets.models import (
    APTAuthenticationData,
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
    AssetCategory,
    BaseAssetDataModel,
    BasicAPTAuthenticationData,
    BasicAPTCredentials,
    CloudProviderAccountData,
    CloudProvidersType,
    DummyProviderAccountData,
    HetznerProviderAccountConfiguration,
    HetznerProviderAccountCredentials,
    HetznerProviderAccountData,
    KeyPurpose,
    SigningKeyData,
    asset_categories,
    asset_data_model,
)

__all__ = [
    "APTAuthenticationData",
    "AWSProviderAccountConfiguration",
    "AWSProviderAccountCredentials",
    "AWSProviderAccountData",
    "AssetCategory",
    "BaseAssetDataModel",
    "BasicAPTAuthenticationData",
    "BasicAPTCredentials",
    "CloudProviderAccountData",
    "CloudProvidersType",
    "DummyProviderAccountData",
    "HetznerProviderAccountConfiguration",
    "HetznerProviderAccountCredentials",
    "HetznerProviderAccountData",
    "KeyPurpose",
    "SigningKeyData",
    "asset_categories",
    "asset_data_model",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Input fields for server tasks."""

from typing import override

from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.tasks.inputs import CollectionInput
from debusine.db.tasks.models import DBInputLookupResult
from debusine.db.tasks.resolver import DBFieldResolver
from debusine.server.tasks.inputs.models import CopyCollectionItemsCopiesInfo
from debusine.server.tasks.models import CopyCollectionItemsCopies
from debusine.tasks import TaskConfigError
from debusine.tasks.inputs import FieldResolver, serializers
from debusine.tasks.inputs.inputs import DataFieldInput


class CopyCollectionItemsCopiesInput(
    DataFieldInput[
        list[CopyCollectionItemsCopies], list[CopyCollectionItemsCopiesInfo]
    ]
):
    """A set of copies for the CopyCollectionItems task."""

    serializer = serializers.PydanticListSerializer(
        CopyCollectionItemsCopiesInfo
    )

    @override
    def resolve(
        self, resolver: FieldResolver
    ) -> list[CopyCollectionItemsCopiesInfo]:
        assert isinstance(resolver, DBFieldResolver)
        all_copies = resolver.get_field(self)
        assert all_copies is not None
        info: list[CopyCollectionItemsCopiesInfo] = []
        for i, copies in enumerate(all_copies):
            field_prefix = f"{self.field}[{i}]"
            source_items_field_name = f"{field_prefix}.source_items"
            try:
                source_items_results = resolver.work_request.lookup_multiple(
                    lookup=copies.source_items, expect_type=LookupChildType.ANY
                )
            except LookupError as e:
                raise TaskConfigError(
                    f"Lookup for field '{source_items_field_name}' in"
                    f" {resolver.work_request.workspace} failed: {e}"
                ) from e
            source_items: list[DBInputLookupResult] = []
            for result in source_items_results:
                if result.result_type == LookupResultType.COLLECTION:
                    raise TaskConfigError(
                        f"Bad lookup result from field"
                        f" '{source_items_field_name}' in"
                        f" {resolver.work_request.workspace}:"
                        f" Cannot copy entire collections"
                    )
                source_items.append(
                    DBInputLookupResult.from_lookup_result(
                        result, label=source_items_field_name
                    )
                )
            target_collection_field = CollectionInput(
                field=f"{field_prefix}.target_collection"
            )
            target_collection_field.getter = lambda _: copies.target_collection
            info.append(
                CopyCollectionItemsCopiesInfo(
                    label=field_prefix,
                    source_items=source_items,
                    target_collection=target_collection_field.resolve(resolver),
                    unembargo=copies.unembargo,
                    replace=copies.replace,
                    name_template=copies.name_template,
                    variables=copies.variables,
                )
            )

        return info

    @override
    def get_artifact_ids(
        self, value: list[CopyCollectionItemsCopiesInfo]
    ) -> list[int]:
        ids: list[int] = []
        for copies in value:
            ids += [
                item.result.artifact.id
                for item in copies.source_items
                if item.result.artifact is not None
            ]
        return ids



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the lookup views."""

from abc import ABC, abstractmethod
from typing import Any, ClassVar, override

from django.urls import reverse
from rest_framework import status

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.client.models import (
    LookupChildType,
    LookupMultipleResponse,
    LookupResultType,
    LookupSingleResponse,
)
from debusine.db.models import (
    CollectionItem,
    Token,
    User,
    WorkRequest,
    Workspace,
)
from debusine.server.views.base import BaseAPIView
from debusine.server.views.lookups import LookupMultipleView, LookupSingleView
from debusine.server.views.rest import IsUserTokenAuthenticated
from debusine.tasks.models import LookupMultiple
from debusine.test.django import TestCase, TestResponseType


class LookupTestCase[LookupType](TestCase, ABC):
    """Common tests for Single and Multiple Lookups."""

    user: ClassVar[User]
    token: ClassVar[Token]
    workspace: ClassVar[Workspace]
    view: ClassVar[type[BaseAPIView]]

    # We need at least one abstract method to avoid pytest running this ABC
    # directly.
    # trixie's coverage thinks this needs to be tested
    @property
    @abstractmethod
    def api_endpoint(self) -> str: ...  # pragma: no cover

    @override
    def __init_subclass__(cls, /, view: type[BaseAPIView]) -> None:
        """Configure LookupTestCase."""
        cls.view = view

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.workspace = cls.playground.create_workspace(
            name="public", public=True
        )
        cls.user = cls.playground.get_default_user()
        cls.token = cls.playground.create_user_token()

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        self.assertIn(IsUserTokenAuthenticated, self.view.permission_classes)

        response = self.client.post(
            self.api_endpoint,
            data={
                "lookup": "",
                "workspace": self.workspace.name,
                "expect_type": LookupChildType.ARTIFACT,
            },
            content_type="application/json",
        )

        self.assertEqual(
            response.json(),
            {
                "title": "Error",
                "detail": "Authentication credentials were not provided.",
            },
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_workspace_not_found(self) -> None:
        """The work request must exist."""
        response = self.client.post(
            self.api_endpoint,
            data={
                "lookup": "x",
                "workspace": "unknown",
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            "Error",
            detail_pattern="Object with name=unknown does not exist.",
        )

    def test_workspace_not_visible(self) -> None:
        """The workspace must be visible to the user."""
        workspace = self.playground.create_workspace(name="private")

        response = self.client.post(
            self.api_endpoint,
            data={
                "lookup": "x",
                "workspace": workspace.name,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            f"{self.user} cannot display workspace {workspace}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_workflow_root_not_found(self) -> None:
        """The workflow_root must exist, if specified."""
        response = self.client.post(
            self.api_endpoint,
            data={
                "lookup": "x",
                "workspace": self.workspace.name,
                "workflow_root": -1,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response, "Error", detail_pattern="Invalid pk"
        )

    def test_workflow_root_not_visible(self) -> None:
        """The workflow_root must be visible to the user, if specified."""
        private = self.playground.create_workspace(name="private")
        work_request = self.playground.create_worker_task(workspace=private)

        response = self.client.post(
            self.api_endpoint,
            data={
                "lookup": "x",
                "workspace": self.workspace.name,
                "workflow_root": work_request.id,
                "expect_type": LookupChildType.ARTIFACT,
            },
            headers={"token": self.token.key},
            content_type="application/json",
        )

        self.assertResponseProblem(
            response,
            f"{self.user} cannot display work request {work_request}",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def post_lookup(
        self,
        lookup: LookupType,
        expect_type: LookupChildType,
        default_category: CollectionCategory | None = None,
        workflow_root: WorkRequest | None = None,
    ) -> TestResponseType:
        """Make a lookup request."""
        data = {
            "lookup": lookup,
            "workspace": self.workspace.name,
            "expect_type": expect_type,
            "default_category": default_category,
        }
        if workflow_root:
            data["workflow_root"] = workflow_root.id
        return self.client.post(
            self.api_endpoint,
            data=data,
            headers={"token": self.token.key},
            content_type="application/json",
        )


class LookupSingleViewTests(
    LookupTestCase[int | str],
    view=LookupSingleView,
):
    """Tests for LookupSingleView."""

    @override
    @property
    def api_endpoint(self) -> str:
        return reverse("api:lookup-single")

    def test_success_string(self) -> None:
        """A string lookup succeeds and returns an artifact ID."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]

        response = self.post_lookup(
            f"{artifact.id}@artifacts", LookupChildType.ARTIFACT
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.ARTIFACT,
                "collection_item": None,
                "artifact": artifact.id,
                "collection": None,
                "parent_collection_lookup": None,
                "links": {
                    "self": reverse(
                        "api:artifact", kwargs={"artifact_id": artifact.id}
                    ),
                    "webui_self": artifact.get_absolute_url(),
                },
            },
        )

    def test_success_integer(self) -> None:
        """An integer lookup succeeds and returns an artifact ID."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]

        response = self.post_lookup(artifact.id, LookupChildType.ARTIFACT)

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.ARTIFACT,
                "collection_item": None,
                "artifact": artifact.id,
                "collection": None,
                "parent_collection_lookup": None,
                "links": {
                    "self": reverse(
                        "api:artifact", kwargs={"artifact_id": artifact.id}
                    ),
                    "webui_self": artifact.get_absolute_url(),
                },
            },
        )

    def test_key_error(self) -> None:
        """The view returns an error if the lookup returns no items."""
        response = self.post_lookup(
            "nonsense@artifacts", LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "No matches",
            detail_pattern="'nonsense@artifacts' does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_lookup_error(self) -> None:
        """The view returns an error if the lookup is invalid."""
        response = self.post_lookup(
            "internal@collections", LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "internal@collections is only valid in the context of a "
                "workflow"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_wrong_workspace(self) -> None:
        """The item must be in a visible workspace."""
        workspace = self.playground.create_workspace(name="test")
        artifact = self.playground.create_artifact(workspace=workspace)[0]
        lookup = f"{artifact.id}@artifacts"

        response = self.post_lookup(lookup, LookupChildType.ARTIFACT)

        self.assertResponseProblem(
            response,
            "No matches",
            detail_pattern=f"{lookup!r} does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_wrong_type(self) -> None:
        """The item must be of the correct type."""
        artifact = self.playground.create_artifact(workspace=self.workspace)[0]
        lookup = f"{artifact.id}@artifacts"

        response = self.post_lookup(lookup, LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=f"{lookup!r} is of type 'artifact'"
            " instead of expected 'collection'",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_default_category(self) -> None:
        """The client can request a default category."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )

        response = self.post_lookup("collection", LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "'collection' does not specify a category and the context "
                "does not supply a default"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        response = self.post_lookup(
            "collection",
            LookupChildType.COLLECTION,
            default_category=CollectionCategory.TEST,
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.COLLECTION,
                "collection_item": None,
                "artifact": None,
                "collection": collection.id,
                "parent_collection_lookup": None,
                "links": {
                    "self": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": collection.id,
                            "workspace": collection.workspace.name,
                        },
                    ),
                    "webui_self": collection.get_absolute_url(),
                },
            },
        )

    def test_workflow_root(self) -> None:
        """Lookup in the internal collection of a workflow."""
        artifact, _ = self.playground.create_artifact(workspace=self.workspace)
        workflow = self.playground.create_workflow(
            internal_collection_artifacts=[
                (
                    artifact,
                    {
                        "name": "test",
                        "variables": {"data-key": "value"},
                    },
                )
            ],
            workspace=self.workspace,
        )
        collection = workflow.internal_collection
        assert collection is not None
        collection_item = collection.manager.lookup("name:test")
        assert collection_item
        assert collection_item.data

        response = self.post_lookup(
            "internal@collections/name:test",
            LookupChildType.ARTIFACT,
            workflow_root=workflow,
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            {
                "result_type": LookupResultType.ARTIFACT,
                "collection_item": {
                    "name": collection_item.name,
                    "data": collection_item.data,
                },
                "artifact": artifact.id,
                "collection": None,
                "parent_collection_lookup": "internal@collections",
                "links": {
                    "self": reverse(
                        "api:artifact", kwargs={"artifact_id": artifact.id}
                    ),
                    "webui_self": artifact.get_absolute_url(),
                    "parent_collection": reverse(
                        "api:collection-detail",
                        kwargs={
                            "pk": collection.pk,
                            "workspace": collection.workspace.name,
                        },
                    ),
                    "webui_parent_collection": collection.get_absolute_url(),
                },
            },
        )
        LookupSingleResponse.model_validate(data)


class LookupMultipleViewTests(
    LookupTestCase[dict[str, Any] | list[str | dict[str, Any]]],
    view=LookupMultipleView,
):
    """Tests for LookupMultipleView."""

    @override
    @property
    def api_endpoint(self) -> str:
        return reverse("api:lookup-multiple")

    def test_success(self) -> None:
        """The lookup succeeds and returns an artifact ID."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifacts = [
            self.playground.create_artifact(category=ArtifactCategory.TEST)[0]
            for _ in range(2)
        ]
        collection_items = [
            CollectionItem.objects.create(
                parent_collection=collection,
                name=f"artifact{i}",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.TEST,
                artifact=artifact,
                data={"item-id": i},
                created_by_user=self.playground.get_default_user(),
            )
            for i, artifact in enumerate(artifacts)
        ]

        response = self.post_lookup(
            {"collection": f"collection@{CollectionCategory.TEST}"},
            LookupChildType.ARTIFACT,
        )

        data = self.assertAPIResponseOk(response)
        self.assertCountEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": {
                        "name": collection_item.name,
                        "data": collection_item.data,
                    },
                    "artifact": artifact.id,
                    "collection": None,
                    "parent_collection_lookup": str(collection),
                    "links": {
                        "self": reverse(
                            "api:artifact", kwargs={"artifact_id": artifact.id}
                        ),
                        "webui_self": artifact.get_absolute_url(),
                        "parent_collection": reverse(
                            "api:collection-detail",
                            kwargs={
                                "pk": collection.pk,
                                "workspace": collection.workspace.name,
                            },
                        ),
                        "webui_parent_collection": (
                            collection.get_absolute_url()
                        ),
                    },
                }
                for artifact, collection_item in zip(
                    artifacts, collection_items
                )
            ],
        )
        LookupMultipleResponse.model_validate(data)

    def test_normalized(self) -> None:
        """The view accepts fully-normalized lookups."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifacts = [
            self.playground.create_artifact(category=ArtifactCategory.TEST)[0]
            for _ in range(2)
        ]
        collection_items = [
            CollectionItem.objects.create(
                parent_collection=collection,
                name=f"artifact{i}",
                child_type=CollectionItem.Types.ARTIFACT,
                category=ArtifactCategory.TEST,
                artifact=artifact,
                data={},
                created_by_user=self.playground.get_default_user(),
            )
            for i, artifact in enumerate(artifacts)
        ]

        lookup = LookupMultiple.model_validate(
            {"collection": f"collection@{CollectionCategory.TEST}"}
        ).model_dump()
        self.assertEqual(
            lookup,
            (
                {
                    "category": None,
                    "child_type": "artifact",
                    "collection": f"collection@{CollectionCategory.TEST}",
                    "data_matchers": (),
                    "name_matcher": None,
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
            ),
        )

        response = self.post_lookup(lookup, LookupChildType.ARTIFACT)

        data = self.assertAPIResponseOk(response)
        self.assertCountEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": {
                        "name": collection_item.name,
                        "data": collection_item.data,
                    },
                    "artifact": artifact.id,
                    "collection": None,
                    "parent_collection_lookup": str(collection),
                    "links": {
                        "self": reverse(
                            "api:artifact", kwargs={"artifact_id": artifact.id}
                        ),
                        "webui_self": artifact.get_absolute_url(),
                        "parent_collection": reverse(
                            "api:collection-detail",
                            kwargs={
                                "pk": collection.pk,
                                "workspace": collection.workspace.name,
                            },
                        ),
                        "webui_parent_collection": (
                            collection.get_absolute_url()
                        ),
                    },
                }
                for artifact, collection_item in zip(
                    artifacts, collection_items
                )
            ],
        )

    def test_deserialization_error(self) -> None:
        """The view returns an error if it cannot deserialize the lookup."""
        response = self.post_lookup(
            "nonsense",  # type: ignore[arg-type]
            LookupChildType.ARTIFACT,
        )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.headers["Content-Type"], "application/problem+json"
        )
        self.assertEqual(response.json()["title"], "Cannot deserialize lookup")
        self.assertRegex(
            response.json()["validation_errors"]["lookup"],
            r"^1 validation error for LookupMultiple",
        )

    def test_key_error(self) -> None:
        """The view returns an empty list if the lookup returns no items."""
        response = self.post_lookup(
            {"collection": "nonexistent@test"}, LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "One of the lookups returned no matches",
            detail_pattern="'nonexistent@test' does not exist or is hidden",
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_lookup_error(self) -> None:
        """The view returns an error if the lookup is invalid."""
        response = self.post_lookup(
            {"collection": "internal@collections"},
            LookupChildType.ARTIFACT,
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "internal@collections is only valid in the context of a "
                "workflow"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_wrong_workspace(self) -> None:
        """The item must be in a visible workspace."""
        workspace = self.playground.create_workspace(name="test")
        self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=workspace
        )

        response = self.post_lookup(
            [{"collection": f"collection@{CollectionCategory.TEST}"}],
            LookupChildType.ARTIFACT,
        )

        self.assertResponseProblem(
            response,
            "One of the lookups returned no matches",
            detail_pattern=(
                f"'collection@{CollectionCategory.TEST}' does not exist or is "
                f"hidden"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )

    def test_wrong_type(self) -> None:
        """The lookup must be of the correct type."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        self.playground.create_bare_data_item(collection, "bare")
        lookup = {
            "collection": f"collection@{CollectionCategory.TEST}",
            "child_type": "bare",
        }

        response = self.post_lookup(lookup, LookupChildType.COLLECTION)

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                f"Result from {lookup!r} is of type 'bare' instead of expected"
                f" 'collection'"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    def test_default_category(self) -> None:
        """The client can request a default category."""
        collection = self.playground.create_collection(
            "collection", CollectionCategory.TEST, workspace=self.workspace
        )
        artifact = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )[0]
        collection_item = CollectionItem.objects.create(
            parent_collection=collection,
            name="artifact",
            child_type=CollectionItem.Types.ARTIFACT,
            category=ArtifactCategory.TEST,
            artifact=artifact,
            data={},
            created_by_user=self.playground.get_default_user(),
        )

        response = self.post_lookup(
            {"collection": "collection"}, LookupChildType.ARTIFACT
        )

        self.assertResponseProblem(
            response,
            "Lookup error",
            detail_pattern=(
                "'collection' does not specify a category and the context "
                "does not supply a default"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

        response = self.post_lookup(
            {"collection": "collection"},
            LookupChildType.ARTIFACT,
            default_category=CollectionCategory.TEST,
        )

        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            data,
            [
                {
                    "result_type": LookupResultType.ARTIFACT,
                    "collection_item": {
                        "name": collection_item.name,
                        "data": collection_item.data,
                    },
                    "artifact": artifact.id,
                    "collection": None,
                    "parent_collection_lookup": str(collection),
                    "links": {
                        "self": reverse(
                            "api:artifact", kwargs={"artifact_id": artifact.id}
                        ),
                        "webui_self": artifact.get_absolute_url(),
                        "parent_collection": reverse(
                            "api:collection-detail",
                            kwargs={
                                "pk": collection.pk,
                                "workspace": collection.workspace.name,
                            },
                        ),
                        "webui_parent_collection": (
                            collection.get_absolute_url()
                        ),
                    },
                }
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the task to create child workspaces."""

import datetime as dt
from typing import ClassVar, override

from django.utils import timezone

from debusine.artifacts.models import (
    SINGLETON_COLLECTION_CATEGORIES,
    WorkRequestResults,
)
from debusine.db.models import Group, WorkflowTemplate, Workspace
from debusine.db.playground import scenarios
from debusine.server.tasks import (
    CreateChildWorkspace,
    ServerTaskPermissionDenied,
)
from debusine.server.tasks.create_child_workspace import CannotCreate
from debusine.server.tasks.models import CreateChildWorkspaceData
from debusine.server.tasks.tests.helpers import ServerTaskTestCase
from debusine.test.django import AllowAll, DenyAll, override_permission


class CreateChildWorkspaceTests(ServerTaskTestCase[CreateChildWorkspace]):
    """Tests for :py:class:`CreateChildWorkspace`."""

    SAMPLE_TASK_DATA = {"prefix": "System", "suffix": "test"}

    scenario = scenarios.DefaultContext()
    workflow_template_1: ClassVar[WorkflowTemplate]
    workflow_template_2: ClassVar[WorkflowTemplate]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.workflow_template_1 = cls.playground.create_workflow_template(
            name="workflow_template_1", task_name="noop"
        )
        cls.workflow_template_2 = cls.playground.create_workflow_template(
            name="workflow_template_2", task_name="noop"
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_get_label(self) -> None:
        """Test get_label."""
        self.assertEqual(
            self.task.get_label(), "create child workspace 'System-test'"
        )

    def test_validate_prefix(self) -> None:
        """Test prefix validation."""
        for name in ["test-test"]:
            with self.subTest(name=name):
                CreateChildWorkspaceData(prefix=name, suffix="suffix")

        for name in ["0test", "test@test", ".test", "+test"]:
            with (
                self.subTest(name=name),
                self.assertRaisesRegex(
                    ValueError, r"\bprefix\n.*String should match pattern"
                ),
            ):
                CreateChildWorkspaceData(prefix=name, suffix="suffix")

    def test_validate_suffix(self) -> None:
        """Test suffix validation."""
        for name in ["test-test"]:
            with self.subTest(name=name):
                CreateChildWorkspaceData(prefix="prefix", suffix=name)

        for name in ["0test", "test@test", ".test", "+test"]:
            with (
                self.subTest(name=name),
                self.assertRaisesRegex(
                    ValueError, r"\bsuffix\n.*String should match pattern"
                ),
            ):
                CreateChildWorkspaceData(prefix="prefix", suffix=name)

    @override_permission(Workspace, "can_create_child_workspace", DenyAll)
    def test_execute_no_perms(self) -> None:
        """Test running the task with insufficient permissions."""
        with self.assertRaisesRegex(
            ServerTaskPermissionDenied,
            "playground cannot create a child workspace from debusine/System",
        ):
            self.task.execute()

    @override_permission(Workspace, "can_create_child_workspace", AllowAll)
    def test_execute_defaults(self) -> None:
        """Test running the task with default arguments."""
        test_start = timezone.now()
        self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        new_workspace = Workspace.objects.get(
            scope=self.scenario.scope, name="System-test"
        )
        self.assertTrue(new_workspace.public)
        self.assertEqual(
            new_workspace.default_expiration_delay, dt.timedelta(0)
        )
        self.assertQuerySetEqual(
            new_workspace.inherits.all(), [self.scenario.workspace]
        )
        self.assertGreaterEqual(new_workspace.created_at, test_start)
        self.assertEqual(new_workspace.expiration_delay, dt.timedelta(days=60))

        self.assertQuerySetEqual(
            new_workspace.collections.values_list("category", flat=True),
            SINGLETON_COLLECTION_CATEGORIES,
            ordered=False,
        )

        group = Group.objects.get(
            scope=self.scenario.scope, workspace=new_workspace, name="Owners"
        )
        self.assertTrue(group.ephemeral)
        self.assertQuerySetEqual(group.users.all(), [self.scenario.user])
        # TODO: assert that user is admin of the group (after !1611)

        role = new_workspace.roles.get(group=group)
        self.assertEqual(role.role, Workspace.Roles.OWNER)

        self.assertQuerySetEqual(
            WorkflowTemplate.objects.filter(workspace=new_workspace), []
        )

    @override_permission(Workspace, "can_create_child_workspace", AllowAll)
    def test_execute_all_set(self) -> None:
        """Test running the task with all arguments set."""
        test_start = timezone.now()
        self.task_data["public"] = False
        self.task_data["owner_group"] = self.scenario.workspace_owners.name
        self.task_data["workflow_template_names"] = ["workflow_template_1"]
        self.task_data["expiration_delay"] = 7
        self.assertEqual(self.task.execute(), WorkRequestResults.SUCCESS)

        new_workspace = Workspace.objects.get(
            scope=self.scenario.scope, name="System-test"
        )
        self.assertFalse(new_workspace.public)
        self.assertEqual(
            new_workspace.default_expiration_delay, dt.timedelta(0)
        )
        self.assertQuerySetEqual(
            new_workspace.inherits.all(), [self.scenario.workspace]
        )
        self.assertGreaterEqual(new_workspace.created_at, test_start)
        self.assertEqual(new_workspace.expiration_delay, dt.timedelta(days=7))

        role = new_workspace.roles.get(group=self.scenario.workspace_owners)
        self.assertEqual(role.role, Workspace.Roles.OWNER)

        new_workflow_templates = list(
            WorkflowTemplate.objects.filter(workspace=new_workspace)
        )
        self.assertEqual(len(new_workflow_templates), 1)

        self.assertEqual(
            new_workflow_templates[0].name, self.workflow_template_1.name
        )
        self.assertEqual(new_workflow_templates[0].workspace, new_workspace)
        self.assertEqual(
            new_workflow_templates[0].task_name,
            self.workflow_template_1.task_name,
        )
        self.assertEqual(
            new_workflow_templates[0].static_parameters,
            self.workflow_template_1.static_parameters,
        )
        self.assertEqual(
            new_workflow_templates[0].priority,
            self.workflow_template_1.priority,
        )

    @override_permission(Workspace, "can_create_child_workspace", AllowAll)
    def test_wrong_group_name(self) -> None:
        """Test running the task with a misspelled group name."""
        self.task_data["owner_group"] = "misspelled-name"
        with self.assertRaisesRegex(
            CannotCreate, "owner group misspelled-name not found"
        ):
            self.task.execute()

    @override_permission(Workspace, "can_create_child_workspace", AllowAll)
    def test_wrong_workflowtemplate_name(self) -> None:
        """Test running the task with a misspelled workflow template name."""
        self.task_data["workflow_template_names"] = ["misspelled-name"]
        with self.assertRaisesRegex(
            CannotCreate, "System/misspelled-name does not exist"
        ):
            self.task.execute()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the workspace views."""

from typing import Any, assert_never

from django.urls import reverse
from rest_framework import status

from debusine.client.models import (
    WorkspaceInheritanceChain,
    WorkspaceInheritanceChainElement,
)
from debusine.db.models import Workspace
from debusine.db.models.workspaces import WorkspaceChain
from debusine.db.playground import scenarios
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)


class WorkspaceInheritanceViewTests(TestCase):
    """Tests for :py:class:`WorkspaceInheritanceView`."""

    scenario = scenarios.DefaultContextAPI()

    def get(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
    ) -> TestResponseType:
        """GET request for a workspace."""
        match workspace:
            case None:
                workspace_name = self.scenario.workspace.name
            case str():
                workspace_name = workspace
            case _:
                workspace_name = workspace.name

        match user_token:
            case True:
                headers = {"token": self.scenario.user_token.key}
            case False:
                headers = {}
            case _ as unreachable:
                assert_never(unreachable)

        return self.client.get(
            reverse(
                "api:workspace-inheritance",
                kwargs={"workspace": workspace_name},
            ),
            content_type="application/json",
            headers=headers,
        )

    def post(
        self,
        workspace: Workspace | str | None = None,
        user_token: bool = True,
        data: dict[str, Any] | None = None,
    ) -> TestResponseType:
        """POST request for a named collection."""
        match workspace:
            case None:
                workspace_name = self.scenario.workspace.name
            case str():
                workspace_name = workspace
            case _:
                workspace_name = workspace.name

        match user_token:
            case True:
                headers = {"token": self.scenario.user_token.key}
            case False:
                headers = {}
            case _ as unreachable:
                assert_never(unreachable)

        if data is None:
            data = {}

        return self.client.post(
            reverse(
                "api:workspace-inheritance",
                kwargs={"workspace": workspace_name},
            ),
            data=data,
            content_type="application/json",
            headers=headers,
        )

    def test_unauthenticated(self) -> None:
        """Authentication is required."""
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(user_token=False)
                self.assertResponseProblem(
                    response,
                    "Error",
                    detail_pattern=(
                        "Authentication credentials were not provided."
                    ),
                    status_code=status.HTTP_403_FORBIDDEN,
                )

    def test_workspace_not_found(self) -> None:
        """Workspace must exist."""
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(workspace="does-not-exist")
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace does-not-exist not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_workspace_not_accessible(self) -> None:
        """User must be able to display the workspace."""
        workspace = self.playground.create_workspace(name="private")
        for method in "get", "post":
            with self.subTest(method=method):
                response = getattr(self, method)(workspace=workspace)
                self.assertResponseProblem(
                    response,
                    "Workspace not found",
                    detail_pattern=(
                        "Workspace private not found in scope debusine"
                    ),
                    status_code=status.HTTP_404_NOT_FOUND,
                )

    def test_post_needs_can_configure(self) -> None:
        """The user must be able to display the collection."""
        with override_permission(Workspace, "can_configure", DenyAll):
            response = self.post()
            self.assertResponseProblem(
                response,
                "playground cannot configure workspace debusine/System",
                status_code=status.HTTP_403_FORBIDDEN,
            )

    def test_get_empty_chain(self) -> None:
        response = self.get()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            WorkspaceInheritanceChain(**data),
            WorkspaceInheritanceChain(chain=[]),
        )

    def test_get_chain(self) -> None:
        first = self.playground.create_workspace(name="first")
        second = self.playground.create_workspace(name="second")
        self.scenario.workspace.set_inheritance([first, second])
        response = self.get()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            WorkspaceInheritanceChain(**data),
            WorkspaceInheritanceChain(
                chain=[
                    WorkspaceInheritanceChainElement(
                        id=first.pk, scope="debusine", workspace="first"
                    ),
                    WorkspaceInheritanceChainElement(
                        id=second.pk,
                        scope="debusine",
                        workspace="second",
                    ),
                ]
            ),
        )

        # Try again after swapping order, to see if order is preserved
        chain = WorkspaceChain.objects.get(
            parent=first, child=self.scenario.workspace
        )
        chain.order = 10
        chain.save()
        response = self.get()
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            WorkspaceInheritanceChain(**data),
            WorkspaceInheritanceChain(
                chain=[
                    WorkspaceInheritanceChainElement(
                        id=second.pk,
                        scope="debusine",
                        workspace="second",
                    ),
                    WorkspaceInheritanceChainElement(
                        id=first.pk, scope="debusine", workspace="first"
                    ),
                ]
            ),
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_set_chain(self) -> None:
        first = self.playground.create_workspace(name="first", public=True)
        second = self.playground.create_workspace(name="second", public=True)
        response = self.post(
            data={
                "chain": [
                    {"id": first.pk},
                    {"scope": "debusine", "workspace": second.name},
                ]
            }
        )
        data = self.assertAPIResponseOk(response)
        self.assertEqual(
            WorkspaceInheritanceChain(**data),
            WorkspaceInheritanceChain(
                chain=[
                    WorkspaceInheritanceChainElement(
                        id=first.pk, scope="debusine", workspace="first"
                    ),
                    WorkspaceInheritanceChainElement(
                        id=second.pk,
                        scope="debusine",
                        workspace="second",
                    ),
                ]
            ),
        )

    @override_permission(Workspace, "can_display", AllowAll)
    @override_permission(Workspace, "can_configure", AllowAll)
    def test_set_chain_private_workspace_in_other_scope(self) -> None:
        scope = self.playground.get_or_create_scope("otherscope")
        second = self.playground.create_workspace(name="other", scope=scope)
        self.playground.create_group_role(
            second, Workspace.Roles.OWNER, [self.scenario.user]
        )

        response = self.post(data={"chain": [{"id": second.pk}]})
        self.assertResponseProblem(
            response,
            "Workspace cannot be inherited",
            detail_pattern=(
                "Private workspaces from other scopes cannot be inherited."
            ),
            status_code=status.HTTP_403_FORBIDDEN,
        )

    @override_permission(Workspace, "can_configure", AllowAll)
    def test_set_invalid_chain(self) -> None:
        first = self.playground.create_workspace(name="first", public=True)
        response = self.post(
            data={
                "chain": [
                    {"id": first.pk},
                    {"id": first.pk},
                ]
            }
        )
        self.assertResponseProblem(
            response,
            "Invalid inheritance chain",
            detail_pattern="duplicate workspace 'first' in inheritance chain",
            status_code=status.HTTP_400_BAD_REQUEST,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View extension to display contents of files."""

import io
import mimetypes
import mmap
import os.path
from collections.abc import Generator, Iterable
from functools import cached_property
from pathlib import Path
from typing import (
    Annotated,
    Any,
    NamedTuple,
    Self,
    TYPE_CHECKING,
    TypeAlias,
    override,
)

import pydantic
import pygments
import pygments.formatters
import pygments.lexers
from django.http import FileResponse
from django.http.response import HttpResponseBase
from django.shortcuts import redirect
from django.utils.html import escape, format_html
from django.utils.http import parse_header_parameters
from django.utils.safestring import SafeString
from django.views.generic.base import View
from pydantic_core import core_schema
from rest_framework import status

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import FileInArtifact
from debusine.server.views import ProblemResponse
from debusine.utils import parse_range_header
from debusine.web.views import view_utils
from debusine.web.views.base import TemplateWidget

if TYPE_CHECKING:
    from _typeshed import SupportsRead

    SupportsRead  # fake usage for vulture
    HtmlFormatter: TypeAlias = pygments.formatters.HtmlFormatter
else:
    # pygments doesn't support generic types at run-time yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    HtmlFormatter = patch_runtime_generic_support(
        pygments.formatters.HtmlFormatter
    )


# Above this file size, only offer to view or download raw
MAX_FILE_SIZE = 2 * 1024 * 1024


class FileUI(NamedTuple):
    """Information about how to display a file."""

    #: content type to use for downloads
    content_type: str
    #: tag identifying specialised view functions to display the file
    widget_class: type["FileWidget"]
    #: pygments lexer to use (default: autodetect from mimetype)
    pygments_lexer: str | None = None

    @classmethod
    def from_file_in_artifact(cls, file: FileInArtifact) -> Self:
        """Get a FileUI for a file."""
        content_type: str
        widget_class: type[FileWidget]

        if (
            file.artifact.category == ArtifactCategory.AUTOPKGTEST
            and file.path == "log"
        ):
            content_type = "text/plain; charset=utf-8"
            widget_class = LogFileWidget
        elif file.content_type is not None:
            content_type = file.content_type
            main_type, _ = parse_header_parameters(content_type)
            if main_type.startswith("text/") or main_type in {
                "application/json",
                "application/pgp-signature",
                "inode/x-empty",
            }:
                if main_type not in {
                    "text/plain",
                    "text/pgp",  # lower-cased by parse_header_parameters
                    "text/markdown",
                    "application/json",
                    "application/pgp-signature",
                }:
                    # Serve other text types as text/plain; clients should
                    # not be allowed to tell Debusine to serve arbitrary
                    # JavaScript or similar.
                    content_type = "text/plain"
                widget_class = TextFileWidget
            else:
                if main_type not in {
                    "application/gzip",
                    "application/octet-stream",
                    "application/vnd.debian.binary-package",
                    "application/x-bzip2",
                    "application/x-tar",
                    "application/x-xz",
                    "application/zstd",
                }:
                    # Serve other binary types as application/octet-stream,
                    # to avoid attacks where clients try to tell Debusine to
                    # serve something like malicious images that browsers
                    # may try to interpret.
                    content_type = "application/octet-stream"

                widget_class = BinaryFileWidget
        else:
            match os.path.splitext(file.path)[1]:
                case ".buildinfo" | ".dsc" | ".changes":
                    content_type = "text/plain; charset=utf-8"
                    widget_class = TextFileWidget
                case ".txt" | ".log" | ".build" | ".buildlog" | ".sources":
                    content_type = "text/plain; charset=utf-8"
                    widget_class = TextFileWidget
                case '.md':
                    content_type = "text/markdown; charset=utf-8"
                    widget_class = TextFileWidget
                case _:
                    # Logic taken from django FileResponse.set_headers
                    encoding_map = {
                        'bzip2': 'application/x-bzip',
                        'gzip': 'application/gzip',
                        'xz': 'application/x-xz',
                    }
                    _content_type, encoding = mimetypes.guess_file_type(
                        file.path
                    )
                    content_type = _content_type or 'application/octet-stream'
                    # Encoding isn't set to prevent browsers from automatically
                    # uncompressing files.
                    if encoding:
                        content_type = encoding_map.get(encoding, content_type)
                    widget_class = BinaryFileWidget

        pygments_lexer: str | None = None
        match os.path.splitext(file.path)[1]:
            case ".buildinfo" | ".dsc" | ".changes":
                pygments_lexer = "debcontrol"

        if file.file.size > MAX_FILE_SIZE:
            widget_class = TooBigFileWidget

        return cls(content_type, widget_class, pygments_lexer)


def add_line_number(
    *, line_number: int, suffix: str, line: SafeString
) -> SafeString:
    """Return HTML to render the line number and contents."""
    return format_html(
        '<a id="L{line_number}{suffix}" href="#L{line_number}{suffix}">'
        '{line_number}</a>'
        '<i>{line}</i>\n',
        line_number=line_number,
        suffix=suffix,
        line=line,
    )


class LinenoHtmlFormatter(HtmlFormatter[Any]):
    """HtmlFormatter that keeps a count of line numbers."""

    def __init__(self, *args: Any, suffix: str, **kwargs: Any) -> None:
        """Keep a count of line numbers."""
        self.suffix = suffix
        self.linenumber_count = 0
        super().__init__(*args, **kwargs)

    @override
    def wrap(
        self, source: Iterable[tuple[int, str]]
    ) -> Generator[tuple[int, str]]:
        """Format each line with the line number."""
        for formatted, line in source:
            # formatted can be:
            # - 1: line of formatted code
            # - 0: Pygments wrapping
            if formatted:
                self.linenumber_count += 1
                yield (
                    1,
                    add_line_number(
                        line_number=self.linenumber_count,
                        suffix=self.suffix,
                        line=SafeString(line.rstrip("\n")),
                    ),
                )
            else:
                yield 0, line


class FileWidget(TemplateWidget):
    """Widget that displays a file."""

    def __init__(
        self,
        file_in_artifact: FileInArtifact,
        *,
        file_ui: FileUI | None = None,
        file_tag: str | None = None,
        file_contents: bytes | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize object with a file to display.

        :param file_tag: there might be pages with more
          than one FileWidget (even referring to the same file). In order
          to create unique anchors to the lines, it's possible to pass a
          file_tag. The anchors will contain it, currently in the form of:
          "#L{line_number}-{file_tag}"
        :param file_contents: if set and smaller than MAX_FILE_SIZE it will
          be used by read_file() method. Otherwise, it will be read on demand
           by the method read_file().
        """
        self.file_in_artifact = file_in_artifact
        self.file_ui = file_ui or FileUI.from_file_in_artifact(file_in_artifact)
        self.file_tag = file_tag
        self.file_contents = None

        if file_contents is not None and len(file_contents) <= MAX_FILE_SIZE:
            self.file_contents = file_contents

    @classmethod
    def create(
        cls,
        file_in_artifact: FileInArtifact,
        *,
        file_ui: FileUI | None = None,
        file_tag: str | None = None,
        file_contents: bytes | None = None,
        **kwargs: Any,
    ) -> "FileWidget":
        """Create a file widget for the given file."""
        file_ui = file_ui or FileUI.from_file_in_artifact(file_in_artifact)
        return file_ui.widget_class(
            file_in_artifact,
            file_ui=file_ui,
            file_tag=file_tag,
            file_contents=file_contents,
            **kwargs,
        )

    def suffix(self) -> str:
        """Return suffix for the ids: "-{self.file_tag}" or ""."""
        return f"-{self.file_tag}" if self.file_tag else ""

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["file_in_artifact"] = self.file_in_artifact
        ctx["file_ui"] = self.file_ui
        return ctx

    def read_file(self) -> bytes:
        """Return self.file_contents or read it."""
        if self.file_contents is not None:
            return self.file_contents

        self.file_contents = view_utils.read_file(
            self.file_in_artifact, max_file_size=MAX_FILE_SIZE
        )
        return self.file_contents


class LogFileWidget(FileWidget):
    """Show log files."""

    class Section(pydantic.BaseModel):
        """Information from a section of the log.."""

        title: str
        start_line: int
        end_line: int
        expanded: bool

    class RenderedSection(pydantic.BaseModel):
        """Rendered section: content and other required information."""

        start_line: int
        title: str | None
        expanded: bool
        content: Annotated[
            SafeString,
            pydantic.GetPydanticSchema(
                lambda *_: core_schema.is_instance_schema(SafeString)
            ),
        ]

    template_name = "web/_file_log.html"

    def __init__(
        self,
        file_in_artifact: FileInArtifact,
        *,
        sections: list[Section] | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize LogFileWidget."""
        super().__init__(file_in_artifact, **kwargs)

        self._sections = sections if sections is not None else []

        LogFileWidget.RenderedSection.model_rebuild()

    @staticmethod
    def _add_line_numbers(
        lines: list[str], *, start: int, end: int, suffix: str
    ) -> SafeString:
        return SafeString(
            "".join(
                add_line_number(line_number=n, suffix=suffix, line=escape(line))
                for n, line in enumerate(lines[start - 1 : end], start=start)
            )
        )

    def _render(
        self, file_contents: list[str], suffix: str
    ) -> list[RenderedSection]:
        rendered_sections = []
        line: int = 1
        for section in self._sections:
            if line < section.start_line:
                # Some lines of the file are not in any Section. Display them
                # (without title, so are not collapsible)
                rendered_sections.append(
                    LogFileWidget.RenderedSection(
                        start_line=line,
                        title=None,
                        expanded=True,
                        content=self._add_line_numbers(
                            file_contents,
                            start=line,
                            end=section.start_line - 1,
                            suffix=suffix,
                        ),
                    )
                )

            rendered_sections.append(
                LogFileWidget.RenderedSection(
                    start_line=section.start_line,
                    title=section.title,
                    expanded=section.expanded,
                    content=self._add_line_numbers(
                        file_contents,
                        start=section.start_line,
                        end=section.end_line,
                        suffix=suffix,
                    ),
                )
            )

            line = section.end_line + 1

        end_line = len(file_contents)
        if line <= end_line:
            # Add any remaining lines of the file
            rendered_sections.append(
                LogFileWidget.RenderedSection(
                    start_line=line,
                    title=None,
                    expanded=True,
                    content=self._add_line_numbers(
                        file_contents,
                        start=line,
                        end=end_line,
                        suffix=suffix,
                    ),
                )
            )

        return rendered_sections

    @override
    def get_context_data(self) -> dict[str, Any]:
        """Add context data for log files."""
        context = super().get_context_data()

        context["formatted_content_div_id"] = (
            f"formatted-content{self.suffix()}"
        )

        if not self.file_in_artifact.complete:
            return context

        raw_bytes = self.read_file()
        raw_str = raw_bytes.decode("utf-8", errors="replace")
        file_contents = raw_str.replace("\r", "^M")
        file_contents = file_contents.rstrip("\n")

        file_contents_list = file_contents.split("\n")

        context["sections"] = self._render(file_contents_list, self.suffix())
        context["suffix"] = self.suffix()
        context["file_lineno_width"] = max(len(str(len(file_contents_list))), 2)

        return context


class TextFileWidget(FileWidget):
    """Show text files."""

    template_name = "web/_file_text.html"

    # add_context_data_* names match FileViewTypes enum options

    @override
    def get_context_data(self) -> dict[str, Any]:
        """Add context data for text files."""
        context = super().get_context_data()

        context["formatted_content_div_id"] = (
            f"formatted-content{self.suffix()}"
        )

        if self.file_in_artifact.complete:
            # Generate CSS with:
            # python3 -m pygments -S github-dark -f html \
            #     > debusine/web/static/web/css/debusine-code-highlight.css
            # See https://pygments.org/styles/ for a list of styles
            if self.file_ui.pygments_lexer:
                lexer = pygments.lexers.get_lexer_by_name(
                    self.file_ui.pygments_lexer
                )
            else:
                lexer = None
                try:
                    lexer = pygments.lexers.get_lexer_for_mimetype(
                        self.file_ui.content_type
                    )
                except pygments.util.ClassNotFound:
                    # Try using the main_type only
                    main_type, _ = parse_header_parameters(
                        self.file_ui.content_type
                    )
                    try:
                        lexer = pygments.lexers.get_lexer_for_mimetype(
                            main_type
                        )
                    except pygments.util.ClassNotFound:
                        pass

                if lexer is None:
                    lexer = pygments.lexers.get_lexer_for_mimetype("text/plain")
            assert lexer is not None

            formatter = LinenoHtmlFormatter(
                cssclass="formatted-content",
                linenos=False,
                suffix=self.suffix(),
            )

            formatted = pygments.highlight(
                self.read_file().decode(errors="replace"),
                lexer,
                formatter,
            )

            context["file_content"] = SafeString(formatted)
            context["file_lineno_width"] = max(
                len(str(formatter.linenumber_count)), 2
            )

        return context


class BinaryFileWidget(FileWidget):
    """Show binary files."""

    template_name = "web/_file_binary.html"


class TooBigFileWidget(FileWidget):
    """Show files that are too big to be loaded."""

    template_name = "web/_file_too_big.html"

    @override
    def get_context_data(self) -> dict[str, Any]:
        """Add context data for files too big to display."""
        context = super().get_context_data()
        context["file_max_size"] = MAX_FILE_SIZE
        return context


class FileDownloadMixin(View):
    """File streaming functions for views."""

    def stream_file(
        self,
        file_in_artifact: FileInArtifact,
        ui_info: FileUI,
        download: bool = True,
    ) -> HttpResponseBase:
        """Return a response that streams the_given file."""
        if not file_in_artifact.complete:
            return ProblemResponse(
                "Cannot download incomplete file",
                status_code=status.HTTP_404_NOT_FOUND,
            )

        scope = file_in_artifact.artifact.workspace.scope
        file_backend = scope.download_file_backend(file_in_artifact.file)
        entry = file_backend.get_entry(file_in_artifact.file)

        url = entry.get_url(content_type=ui_info.content_type)
        if url is not None:
            # The client can download the file from the backend
            # TODO: this does not allow to set content-disposition, that is, to
            # distinguish between "view raw" or "download". Not sure if it can
            # be solved.
            return redirect(url)

        with entry.get_stream() as file:
            file_size = file_in_artifact.file.size
            try:
                content_range = parse_range_header(
                    self.request.headers, file_size
                )
            except ValueError as exc:
                # It returns ProblemResponse because ranges are not used
                # by end users directly
                return ProblemResponse(str(exc))

            status_code: int
            start: int | None
            end: int | None
            if content_range is None:
                # Whole file
                status_code = status.HTTP_200_OK
                start = 0
                end = file_size - 1
            else:
                # Part of a file
                status_code = status.HTTP_206_PARTIAL_CONTENT
                start = content_range["start"]
                end = content_range["end"]

                # It returns ProblemResponse because ranges are not used
                # by end users directly
                if start > file_size:
                    return ProblemResponse(
                        f"Invalid Content-Range start: {start}. "
                        f"File size: {file_size}"
                    )

                elif end >= file_size:
                    return ProblemResponse(
                        f"Invalid Content-Range end: {end}. "
                        f"File size: {file_size}"
                    )

            # Use mmap:
            # - No support for content-range or file chunk in Django
            #   as of 2023, so create filelike object of the right chunk
            # - Prevents FileResponse.file_to_stream.name from taking
            #   precedence over .filename and break mimestype
            file_partitioned: Iterable[object]
            if file_size == 0:
                # cannot mmap an empty file
                file_partitioned = io.BytesIO(b"")
            else:
                # We have to rely on Django to close the mmap on completion
                file_partitioned = mmap.mmap(
                    file.fileno(), end + 1, prot=mmap.PROT_READ
                )
                file_partitioned.seek(start)

            filename = Path(file_in_artifact.path)

            response = FileResponse(
                file_partitioned,
                filename=filename.name,
                status=status_code,
            )

            response["Accept-Ranges"] = "bytes"
            response["Content-Length"] = end - start + 1
            if file_size > 0:
                response["Content-Range"] = f"bytes {start}-{end}/{file_size}"
            if download:
                disposition = "attachment"
            else:
                disposition = "inline"
            response["Content-Disposition"] = (
                f'{disposition}; filename="{filename.name}"'
            )
            response["Content-Type"] = ui_info.content_type

            return response


class PathMixin(View):
    """View that accepts a path kwarg."""

    @staticmethod
    def normalize_path(path: str) -> str:
        """
        Normalize a path used as a subdirectory prefix.

        It will also constrain paths not to point above themselves by extra ../
        components
        """
        path = os.path.normpath(path).strip("/")
        while path.startswith("../"):
            path = path[3:]
        if path in ("", ".", ".."):
            return "/"
        return f"/{path}/"

    @cached_property
    def path(self) -> str:
        """
        Return the current subdirectory as requested by the user.

        This returns a path relative to the root of the artifact, with ""
        standing for the whole artifact
        """
        if not (path := self.kwargs.get("path")):
            return ""
        return self.normalize_path(path)[1:]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Implementation of Amazon S3 file backend."""

import datetime as dt
import filecmp
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Self, TYPE_CHECKING, override
from urllib.parse import urlsplit

import boto3

from debusine.assets.models import AWSProviderAccountData, CloudProvidersType
from debusine.db.models import File, FileStore
from debusine.server.file_backend.interface import (
    FileBackendEntryInterface,
    FileBackendInterface,
)
from debusine.server.file_backend.models import S3FileBackendConfiguration

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client

    S3Client  # fake usage for vulture


class UNKNOWN:
    """
    Sentinel object unknown mtime.

    Replace with PEP-661 sentinels with python 3.15.
    """


class S3FileBackendEntry(FileBackendEntryInterface["S3FileBackend", str]):
    """
    An entry in an Amazon S3 file backend.

    S3 identifiers are "hash-size", where "hash" is the SHA-256 hex digest
    of the file contents and "size" is the file size in bytes.
    """

    @classmethod
    def _create_from_file(cls, backend: "S3FileBackend", fileobj: File) -> Self:
        """Make an entry based on a `File`."""
        return cls(
            backend=backend, identifier=f"{fileobj.sha256.hex()}-{fileobj.size}"
        )

    @override
    def __init__(
        self,
        backend: "S3FileBackend",
        identifier: str,
        mtime: dt.datetime | type[UNKNOWN] = UNKNOWN,
    ) -> None:
        super().__init__(backend, identifier)
        self._mtime = mtime

    @override
    def __str__(self) -> str:
        """Return a string describing this entry."""
        return self._identifier

    @override
    def get_local_path(self) -> None:
        """Return None: no local paths for S3 files."""
        return None

    @override
    @contextmanager
    def get_temporary_local_path(self) -> Generator[Path]:
        """Yield a temporary local path for the file."""
        with tempfile.NamedTemporaryFile(
            prefix="debusine-s3-", delete_on_close=False
        ) as temp_file:
            self.backend.client.download_fileobj(
                self.backend.configuration.bucket_name,
                self._identifier,
                temp_file,
            )
            temp_file.close()
            yield Path(temp_file.name)

    @override
    def get_url(self, content_type: str | None = None) -> str:
        """Return a presigned URL for the file."""
        params = {
            "Bucket": self.backend.configuration.bucket_name,
            "Key": self._identifier,
        }
        if content_type is not None:
            params["ResponseContentType"] = content_type
        url = self.backend.client.generate_presigned_url(
            "get_object",
            Params=params,
            # Hardcoded to one minute for now.  These URLs are currently
            # returned in redirects which are presumably followed promptly
            # by clients.
            ExpiresIn=60,
        )
        assert isinstance(url, str)
        return url

    @override
    def get_mtime(self) -> dt.datetime:
        if self._mtime is UNKNOWN:
            raise NotImplementedError(
                "Cannot look up modification files for individual files"
            )
        assert isinstance(self._mtime, dt.datetime)
        return self._mtime

    @override
    def add(self, local_path: Path) -> None:
        extra_args = {
            "ChecksumAlgorithm": "SHA256",
        }
        if self.backend.configuration.storage_class is not None:
            extra_args["StorageClass"] = (
                self.backend.configuration.storage_class
            )
        self.backend.client.upload_file(
            str(local_path),
            self.backend.configuration.bucket_name,
            self._identifier,
            ExtraArgs=extra_args,
        )

    @override
    def same_contents(self, local_path: Path) -> bool:
        with self.get_temporary_local_path() as temp_path:
            # This standard library function maintains a cache indexed by
            # the paths and stat signatures (type, size, mtime) of both
            # files.  Since `temp_path` was always created in this method,
            # the cache will never be hit.
            return filecmp.cmp(local_path, temp_path, shallow=False)

    @override
    def remove(self) -> None:
        self.backend.client.delete_object(
            Bucket=self.backend.configuration.bucket_name, Key=self._identifier
        )


class S3FileBackend(FileBackendInterface[S3FileBackendConfiguration]):
    """Amazon S3 file backend."""

    def __init__(self, file_store: FileStore) -> None:
        """Initialize `S3FileBackend`."""
        super().__init__()
        self.db_store = file_store
        if file_store.provider_account is None:
            raise RuntimeError(
                "S3 file backend requires a cloud provider account"
            )
        if (
            file_store.provider_account.data.get("provider_type")
            != CloudProvidersType.AWS
        ):
            raise RuntimeError(
                "S3 file backend requires an AWS cloud provider account"
            )
        provider_account_data = file_store.provider_account.data_model
        assert isinstance(provider_account_data, AWSProviderAccountData)
        s3_endpoint_url = provider_account_data.configuration.s3_endpoint_url
        self._is_hetzner = s3_endpoint_url is not None and urlsplit(
            s3_endpoint_url
        ).netloc.endswith(".your-objectstorage.com")
        self.client = self._make_client(provider_account_data)

    def _make_client(
        self, provider_account_data: AWSProviderAccountData
    ) -> "S3Client":
        """
        Make an S3 client for this provider account.

        This exists mainly so that it can be patched by tests.
        """
        return boto3.client(
            "s3",
            region_name=provider_account_data.configuration.region_name,
            endpoint_url=provider_account_data.configuration.s3_endpoint_url,
            aws_access_key_id=provider_account_data.credentials.access_key_id,
            aws_secret_access_key=(
                provider_account_data.credentials.secret_access_key
            ),
        )

    @override
    def get_entry(self, fileobj: File) -> S3FileBackendEntry:
        return S3FileBackendEntry._create_from_file(self, fileobj)

    @override
    def list_entries(self) -> Generator[S3FileBackendEntry]:
        paginator = self.client.get_paginator("list_objects_v2")
        iterator = paginator.paginate(Bucket=self.configuration.bucket_name)
        for page in iterator:
            for item in page.get("Contents", []):
                yield S3FileBackendEntry(
                    backend=self,
                    identifier=item["Key"],
                    mtime=item["LastModified"],
                )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for DebusineBaseCommand class."""

import signal
from io import StringIO
from typing import Any, override
from unittest import mock

from django.core.management import CommandError
from django.db.utils import DatabaseError

from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.test.django import TestCase, TransactionTestCase


class TestSignalHandlersMixin:
    """Class to set and restore SIGINT and SIGTERM signal handlers."""

    def setUp(self) -> None:
        """Initialize object."""
        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

    def tearDown(self) -> None:
        """Restore signals."""
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)


class DebusineBaseCommandRunFromArgv(
    TestSignalHandlersMixin, TransactionTestCase
):
    """Tests for the DebusineBaseCommand that use run_from_argv method."""

    def test_signal_handlers_sigint_sigterm_setup(self) -> None:
        """Test DebusineBaseCommand set SIGINT/SIGTERM handlers."""

        class DebusineSignalSpy(DebusineBaseCommand):
            @override
            def handle(self, *args: Any, **kwargs: Any) -> None:
                self.sigint_handler = signal.getsignal(signal.SIGINT)
                self.sigterm_handler = signal.getsignal(signal.SIGTERM)

        debusine_command = DebusineSignalSpy()

        # DebusineBaseCommand.__init__ does not change the signal handlers
        self.assertEqual(
            signal.getsignal(signal.SIGINT), self.default_sigint_handler
        )
        self.assertEqual(
            signal.getsignal(signal.SIGTERM), self.default_sigterm_handler
        )

        debusine_command.run_from_argv(["debusine-admin", "test-command"])

        # DebusineBaseCommand.execute() set the handlers correctly (at the time
        # that DebusineSignalSpy.handle() was called the handlers were correctly
        # set)
        self.assertEqual(
            debusine_command.sigint_handler, DebusineBaseCommand._exit_handler
        )
        self.assertEqual(
            debusine_command.sigterm_handler, DebusineBaseCommand._exit_handler
        )

        # It is a global change
        self.assertEqual(
            signal.getsignal(signal.SIGINT), DebusineBaseCommand._exit_handler
        )
        self.assertEqual(
            signal.getsignal(signal.SIGTERM), DebusineBaseCommand._exit_handler
        )

    def test_exit_handler(self) -> None:
        """DebusineBaseCommand.exit_handler() raises SystemExit(3)."""
        with self.assertRaisesSystemExit(3):
            DebusineBaseCommand._exit_handler(signal.SIGINT, None)

    @staticmethod
    def run_set_verbosity_value_in_handle(
        verbosity_option: str, verbosity_level: int
    ) -> int:
        """
        Run a command and set self.verbosity_value to self.verbosity.

        DebusineBaseCommand.execute() is expected to set self.verbosity
        before "handle()" is called. handle() set verbosity_value to allow
        asserting that self.verbosity was set correctly.
        """

        class SetVerbosityValueInHandle(DebusineBaseCommand):
            @override
            def handle(self, *args: Any, **kwargs: Any) -> None:
                self.verbosity_value = self.verbosity

        command = SetVerbosityValueInHandle()
        command.run_from_argv(
            [
                "debusine-admin",
                "test-command",
                verbosity_option,
                str(verbosity_level),
            ]
        )
        assert isinstance(command.verbosity_value, int)
        return command.verbosity_value

    def test_verbosity_long_is_set(self) -> None:
        """DebusineBaseCommand.execute() set verbosity to --verbosity value."""
        verbosity = 3
        self.assertEqual(
            self.run_set_verbosity_value_in_handle("--verbosity", verbosity),
            verbosity,
        )

    def test_verbosity_short_is_set(self) -> None:
        """DebusineBaseCommand.execute() set verbosity to -v value."""
        verbosity = 2
        self.assertEqual(
            self.run_set_verbosity_value_in_handle("-v", verbosity), verbosity
        )


class DebusineBaseCommandTests(TestSignalHandlersMixin, TestCase):
    """Tests for the DebusineBaseCommand."""

    def test_operational_error_is_caught(self) -> None:
        """Test that OperationalError raised in execute is dealt with."""
        patch = mock.patch(
            "django.core.management.BaseCommand.execute",
            side_effect=DatabaseError,
        )
        patch.start()
        self.addCleanup(patch.stop)

        err = StringIO()
        command = DebusineBaseCommand(stderr=err)

        with self.assertRaisesRegex(CommandError, "^Database error: ") as exc:
            command.execute(verbosity="1")

        self.assertEqual(exc.exception.returncode, 3)

    def test_print_verbose(self) -> None:
        """Test print_verbose() when used with different verbosity levels."""
        stdout = StringIO()
        command = DebusineBaseCommand(stdout=stdout)

        for verbosity, msg, expected in (
            (1, "test", ""),
            (2, "test", "test\n"),
        ):
            with (
                self.subTest(verbosity=verbosity),
                self.spy_object(stdout, "flush") as spy_flush,
            ):
                command.verbosity = verbosity
                command.print_verbose(msg)
                self.assertEqual(stdout.getvalue(), expected)
                if expected:
                    spy_flush.assert_called_once_with()
                else:
                    spy_flush.assert_not_called()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Merge the ``.changes`` files from multiple uploads."""

import re
from collections import defaultdict
from collections.abc import Sequence
from pathlib import Path
from typing import override

from debian.deb822 import Changes

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianUpload,
    WorkRequestResults,
    get_architectures,
    get_source_package_name,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseExternalTask
from debusine.tasks.inputs import MultiInput, Stage
from debusine.tasks.models import MergeUploadsData, MergeUploadsDynamicData
from debusine.utils import (
    find_files_suffixes,
    is_single_path_component,
    read_changes,
)


class MergeUploadsError(Exception):
    """An error raised while merging upload artifacts."""


class MergeUploads(BaseExternalTask[MergeUploadsData, MergeUploadsDynamicData]):
    """
    Combines multiple debian:upload artifacts into a single one.

    This is in preparation for uploading them together.
    """

    TASK_VERSION = 1

    uploads = MultiInput(
        field="input.uploads",
        categories=[ArtifactCategory.UPLOAD],
        stage=Stage.PENDING,
    )

    @override
    def post_init(self) -> None:
        super().post_init()
        self._changes_paths: list[Path] = []
        self._upload_artifact: Upload | None = None

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        for upload_artifact in self.uploads.artifacts:
            assert isinstance(upload_artifact.data, DebianUpload)
            provided.add(
                ttags.ProvenanceProvided.SYSTEM,
                {
                    ttags.SOURCE_PACKAGE_PREFIX
                    + get_source_package_name(upload_artifact.data)
                },
            )

        return provided, required

    @override
    def get_subject(self) -> str | None:
        source_package_names = set()
        for upload_artifact in self.uploads.artifacts:
            assert isinstance(upload_artifact.data, DebianUpload)
            source_package_names.add(
                get_source_package_name(upload_artifact.data)
            )
        source_package_names_sorted = sorted(source_package_names)

        return (
            source_package_names_sorted[0]
            if len(source_package_names_sorted) == 1
            else None
        )

    @override
    def compute_dynamic_data(self) -> MergeUploadsDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: shared source package name across all uploads
        :parameter_summary: subject (list of architectures)
        """
        architectures = set()
        for upload_artifact in self.uploads.artifacts:
            assert isinstance(upload_artifact.data, DebianUpload)
            architectures |= get_architectures(upload_artifact.data)
        architectures_sorted = ", ".join(sorted(architectures))

        subject = self.get_subject()
        if subject is not None:
            parameter_summary = f"{subject} ({architectures_sorted})"
        else:
            parameter_summary = None

        return MergeUploadsDynamicData(
            input_uploads_ids=self.uploads.resolved_ids(),
            subject=subject,
            parameter_summary=parameter_summary,
        )

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Populate work directory with user-specified binary artifact(s)."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")
        assert self.dynamic_data

        for upload_id in self.dynamic_data.input_uploads_ids:
            artifact = self.debusine.artifact_get(
                upload_id, workspace=self.workspace_name
            )
            if artifact.category != ArtifactCategory.UPLOAD:
                self.append_to_log_file(
                    "fetch_input.log",
                    [
                        f"input.uploads points to a "
                        f"{artifact.category}, not the expected "
                        f"{ArtifactCategory.UPLOAD}."
                    ],
                )
                return False
            upload_dir = destination / f"upload-{upload_id}"
            upload_dir.mkdir()
            self.fetch_artifact(upload_id, upload_dir)

        return True

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        r"""
        Find the .changes files to merge.

        Set self._changes_paths to the relevant files.

        :param download_directory: where to find the .dsc file
          (downloaded via fetch_input)

        :return: True if valid files were found
        """
        # Find the files to merge or early exit if not files
        uploads: dict[int, Path] = {}
        for upload_dir in download_directory.iterdir():
            # fetch_input writes nothing else to the download directory.
            assert upload_dir.is_dir() and upload_dir.name.startswith("upload-")
            uploads[int(upload_dir.name[len("upload-") :])] = upload_dir
        changes_paths: list[Path] = []
        for _, upload_dir in sorted(uploads.items()):
            changes_paths += find_files_suffixes(upload_dir, [".changes"])
        self._changes_paths = changes_paths
        # Ensure we've got 1 .changes file per upload, see:
        # debusine.artifacts.local_artifacts.Upload.files_contain_changes
        assert len(self._changes_paths) >= 1

        return True

    @staticmethod
    def _check_simple_fields(all_changes: Sequence[Changes]) -> None:
        """Check whether simple fields in some .changes files are consistent."""
        if all_changes[0]["Format"] != "1.8":
            raise MergeUploadsError(
                f"Unknown .changes format: {all_changes[0]['Format']}"
            )

        for field in ("Format", "Source", "Version"):
            values = [changes[field] for changes in all_changes]
            if len(set(values)) != 1:
                raise MergeUploadsError(
                    f"{field} fields do not match: {values}"
                )

    @staticmethod
    def _check_descriptions(all_changes: Sequence[Changes]) -> None:
        """Check that descriptions in some .changes files are consistent."""
        changes_description_re = re.compile(r"^ ([^ ]+) - (.+)")
        descriptions: dict[str, str] = {}
        for changes in all_changes:
            for description_line in changes.get("Description", "").splitlines():
                if not description_line:
                    continue
                elif m := changes_description_re.match(description_line):
                    name, description = m.groups()
                    if (
                        name in descriptions
                        and descriptions[name] != description
                    ):
                        raise MergeUploadsError(
                            f"Descriptions for {name} do not match: "
                            f"{descriptions[name]!r} != {description!r}"
                        )
                    descriptions[name] = description

    @staticmethod
    def _check_checksums(all_changes: Sequence[Changes]) -> None:
        """Check that checksums in some .changes files are consistent."""
        checksums: dict[str, dict[str, dict[str, str]]] = defaultdict(dict)
        for changes in all_changes:
            for field in changes:
                if field.lower().startswith(
                    "checksums-"
                ) and field.lower() not in {
                    "checksums-sha1",
                    "checksums-sha256",
                }:
                    raise MergeUploadsError(
                        f"Unsupported checksum field: {field}"
                    )

            for field in ("Files", "Checksums-Sha1", "Checksums-Sha256"):
                for checksum in changes.get(field, []):
                    name = checksum["name"]
                    if (
                        name in checksums[field]
                        and checksums[field][name] != checksum
                    ):
                        raise MergeUploadsError(
                            f"Entries in {field} for {name} do not match: "
                            f"{checksums[field][name]} != {checksum}"
                        )
                    checksums[field][name] = checksum

    def _merge_descriptions(self, merged: Changes, to_merge: Changes) -> None:
        """
        Merge Description from ``to_merge`` into ``merged``.

        Works around a bug fixed here:
        https://salsa.debian.org/python-debian-team/python-debian/-/merge_requests/148
        """
        if to_merge.get("Description"):
            if merged.get("Description"):
                for item in to_merge["Description"].splitlines():
                    if item not in merged["Description"].splitlines():
                        merged["Description"] += "\n" + item
            else:
                merged["Description"] = to_merge["Description"]

    def _symlink_files(
        self, merged_dir: Path, to_merge_path: Path, to_merge: Changes
    ) -> None:
        """Symlink files from ``to_merge`` into ``merged_dir``."""
        for file in to_merge.get("Checksums-Sha256", []):
            # Checked by read_changes.
            assert is_single_path_component(file["name"])
            merged_file_path = merged_dir / file["name"]
            # _check_checksums already checked that there are no overlapping
            # file names.
            assert not merged_file_path.exists()
            merged_file_path.symlink_to(to_merge_path.parent / file["name"])

    def merge_changes(
        self, execute_directory: Path, all_changes: Sequence[Changes]
    ) -> Path:
        """Merge some .changes files."""
        self._check_simple_fields(all_changes)
        self._check_descriptions(all_changes)
        self._check_checksums(all_changes)

        merged_dir = execute_directory / "merged"
        merged_dir.mkdir()
        self._symlink_files(merged_dir, self._changes_paths[0], all_changes[0])

        merged = all_changes[0]
        for to_merge_path, to_merge in zip(
            self._changes_paths[1:], all_changes[1:]
        ):
            merged.merge_fields("Binary", to_merge)
            merged.merge_fields("Architecture", to_merge)
            self._merge_descriptions(merged, to_merge)
            self._symlink_files(merged_dir, to_merge_path, to_merge)

            for field in ("Files", "Checksums-Sha1", "Checksums-Sha256"):
                existing = {
                    tuple(checksum.items()) for checksum in merged[field]
                }
                for checksum in to_merge[field]:
                    assert tuple(checksum.items()) not in existing
                    merged[field].append(checksum)

        merged.order_before("Binary", "Source")
        merged.order_before("Description", "Changes")

        # Use the same "multi" suffix convention as mergechanges(1) from
        # devscripts.
        version_without_epoch = re.sub(r"^\d+:", "", merged["Version"])
        merged_changes_path = (
            merged_dir
            / f"{merged['Source']}_{version_without_epoch}_multi.changes"
        )
        with open(merged_changes_path, "w") as f:
            merged.dump(f, text_mode=True)

        return merged_changes_path

    def make_upload_artifact(self, merged_changes_path: Path) -> Upload:
        """Make an Upload artifact from a merged .changes file."""
        return Upload.create(changes_file=merged_changes_path)

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Do the main work of the task."""
        all_changes = [read_changes(path) for path in self._changes_paths]
        merged_changes_path = self.merge_changes(execute_directory, all_changes)
        self._upload_artifact = self.make_upload_artifact(merged_changes_path)
        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Create DebianUpload artifact and relationships."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")
        assert self.dynamic_data
        assert self._upload_artifact is not None

        changes_uploaded = self.debusine.upload_artifact(
            self._upload_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        for input_upload_id in self.dynamic_data.input_uploads_ids:
            self.debusine.relation_create(
                changes_uploaded.id,
                input_upload_id,
                RelationType.EXTENDS,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Redis backend for prometheus values."""

import datetime as dt
import os
from threading import Event, Thread
from typing import Any, override
from urllib.parse import urlsplit

from django.conf import settings
from django.test.utils import TestContextDecorator, override_settings
from redis import Redis

# For testing, a pool of otherwise anonymous FakeRedis instances are made
# available by ID
_fake_redis_pool: dict[int, Redis] = {}


def redis_client() -> Redis:
    """
    Create a redis client for PROMETHEUS_REDIS_URL.

    Configure the redis database via a URL in PROMETHEUS_REDIS_URL of the form
    redis://localhost:6379/0
    """
    url = getattr(settings, "PROMETHEUS_REDIS_URL")
    assert url
    parsed_url = urlsplit(url)
    assert parsed_url.path.startswith("/")
    assert parsed_url.path[1:].isdigit()
    port = parsed_url.port or 6379
    db = int(parsed_url.path[1:])

    if parsed_url.scheme == "fakeredis":
        from fakeredis import FakeRedis

        if db not in _fake_redis_pool:
            _fake_redis_pool[db] = FakeRedis()
        return _fake_redis_pool[db]

    assert parsed_url.scheme == "redis"
    assert parsed_url.hostname
    return Redis(host=parsed_url.hostname, port=port, db=db)


# For each process identifier, a list of keys that should be kept from expiring
_live_metrics: dict[str, set[str]] = {}


def _key_expiry() -> dt.timedelta:
    """Return the configured expiry for multiprocess keys."""
    return dt.timedelta(
        seconds=int(os.environ.get("PROMETHEUS_REDIS_REFRESH_TTL", 20))
    )


class KeepMetricsAliveThread(Thread):
    """A daemon thread that keeps metrics from expiring as long as we live."""

    stop: Event
    identifier: str
    client: Redis

    @override
    def __init__(
        self, identifier: str, client: Redis, *args: Any, **kwargs: Any
    ) -> None:
        self.stop = Event()
        self.identifier = identifier
        self.client = client
        super().__init__(*args, **kwargs)

    def loop_wait(self, delay: float) -> bool:
        """Patchable wait step."""
        return self.stop.wait(delay)

    @override
    def run(self) -> None:
        delay = float(os.environ.get("PROMETHEUS_REDIS_REFRESH_FREQUENCY", 10))
        expiry = _key_expiry()
        while not self.loop_wait(delay):
            for key in _live_metrics[self.identifier]:
                self.client.expire(key, expiry)


_daemon_threads: dict[str, KeepMetricsAliveThread] = {}


def _keep_key_from_expiring(identifier: str, key: str) -> None:
    """Stop key for process identifier from expiring as long as we are alive."""
    _live_metrics.setdefault(identifier, set()).add(key)
    if identifier not in _daemon_threads:
        thread = KeepMetricsAliveThread(
            identifier=identifier, client=redis_client(), daemon=True
        )
        thread.start()
        _daemon_threads[identifier] = thread


def mark_process_dead(identifier: str | int) -> None:
    """Immediately expire all live* metrics for process identifier."""
    thread = _daemon_threads.pop(str(identifier), None)
    if thread is not None:
        thread.stop.set()
        thread.join()

    keys = _live_metrics.pop(str(identifier), None)
    if not keys:
        return
    redis_client().delete(*keys)


class private_fake_redis(TestContextDecorator):
    """Replace our prometheus redis instance with a fakeredis."""

    @override
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.settings_override = override_settings(
            PROMETHEUS_REDIS_URL="fakeredis:///2"
        )

    @override
    def enable(self) -> None:
        self.settings_override.enable()
        self._old_daemon_threads = _daemon_threads.copy()
        _daemon_threads.clear()
        self._live_metrics = _live_metrics.copy()
        _live_metrics.clear()
        fake_client = redis_client()
        fake_client.flushdb()

    @override
    def disable(self) -> None:
        for identifier in list(_daemon_threads):
            mark_process_dead(identifier)
        _daemon_threads.clear()
        _daemon_threads.update(self._old_daemon_threads)
        _live_metrics.clear()
        _live_metrics.update(self._live_metrics)
        self.settings_override.disable()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to copy items into target collections."""

import shutil
import tempfile
from pathlib import Path
from typing import Any, override

from django.conf import settings

from debusine.artifacts.models import BareDataCategory
from debusine.client.models import LookupResultType
from debusine.db.models import (
    Artifact,
    CollectionItem,
    FileInArtifact,
    User,
    WorkRequest,
    Workspace,
)
from debusine.db.tasks.models import (
    DBInputCollectionSingle,
    DBInputLookupResult,
)
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.inputs import CopyCollectionItemsCopiesInput
from debusine.server.tasks.models import CopyCollectionItemsData


class CannotCopy(Exception):
    """A given source item cannot be copied to a given target collection."""


class CopyCollectionItems(BaseServerTask[CopyCollectionItemsData]):
    """Task that copies items into target collections."""

    TASK_VERSION = 1

    copies = CopyCollectionItemsCopiesInput()

    @staticmethod
    def copy_artifact(
        artifact: Artifact, target_workspace: Workspace
    ) -> Artifact:
        """Copy an artifact to a new workspace."""
        new_artifact = Artifact.objects.create(
            category=artifact.category,
            workspace=target_workspace,
            data=artifact.data,
            expiration_delay=artifact.expiration_delay,
            # The new artifact has a fresh created_at, but we keep
            # created_by and created_by_work_request from the old artifact
            # since that seems more likely to be useful.
            created_by=artifact.created_by,
            created_by_work_request=artifact.created_by_work_request,
            original_artifact=artifact,
        )
        for file_in_artifact in artifact.fileinartifact_set.select_related(
            "file"
        ):
            if not file_in_artifact.complete:
                raise CannotCopy(
                    f"Cannot copy incomplete {file_in_artifact.path!r} from "
                    f"artifact ID {artifact.id}"
                )
            file = file_in_artifact.file
            if not target_workspace.scope.download_file_stores(file).exists():
                # We must copy the file contents.
                try:
                    source_file_backend = (
                        artifact.workspace.scope.download_file_backend(file)
                    )
                except IndexError:
                    raise CannotCopy(
                        f"Cannot copy {file_in_artifact.path!r} from artifact "
                        f"ID {artifact.id}: not in any available file store"
                    )
                target_file_backend = (
                    target_workspace.scope.upload_file_backend(file)
                )
                with (
                    source_file_backend.get_stream(file) as source_file,
                    tempfile.NamedTemporaryFile(
                        prefix="file-copying-",
                        dir=settings.DEBUSINE_UPLOAD_DIRECTORY,
                    ) as temp_file,
                ):
                    shutil.copyfileobj(source_file, temp_file)
                    temp_file.flush()
                    target_file_backend.add_file(Path(temp_file.name), file)
            FileInArtifact.objects.create(
                artifact=new_artifact,
                path=file_in_artifact.path,
                file=file,
                complete=True,
                content_type=file_in_artifact.content_type,
            )
        return new_artifact

    def copy_item(
        self,
        source_item: DBInputLookupResult,
        target_collection: DBInputCollectionSingle,
        *,
        unembargo: bool = False,
        replace: bool = False,
        name_template: str | None = None,
        variables: dict[str, Any] | None = None,
        user: User,
        workflow: WorkRequest | None = None,
    ) -> None:
        """Copy a single item into a target collection."""
        source_result = source_item.result
        match source_result.result_type:
            case LookupResultType.BARE:
                assert source_result.collection_item is not None
                source_workspace = (
                    source_result.collection_item.parent_collection.workspace
                )
                reference_data = {}
            case LookupResultType.ARTIFACT:
                assert source_result.artifact is not None
                source_workspace = source_result.artifact.workspace
                reference_data = source_result.artifact.data
            case _ as unreachable:
                # LookupResultType.COLLECTION is excluded by
                # CopyCollectionItemsCopiesInput.resolve.
                raise AssertionError(
                    f"Unexpected lookup result type: {unreachable}"
                )

        target_workspace = target_collection.collection.workspace

        if not unembargo:
            if not source_workspace.public and target_workspace.public:
                raise CannotCopy(
                    f"Copying from {source_workspace} to {target_workspace} "
                    f"requires unembargo=True"
                )

        merged_variables = (
            source_result.collection_item.data
            if source_result.collection_item is not None
            else {}
        ) | (variables or {})
        try:
            expanded_variables = CollectionItem.expand_variables(
                merged_variables, reference_data
            )
        except (KeyError, ValueError):
            raise CannotCopy(f"Cannot expand variables: {variables}")

        if name_template is not None:
            item_name = CollectionItem.expand_name(
                name_template, expanded_variables
            )
            item_variables = None
        else:
            item_name = None
            item_variables = expanded_variables

        match source_result.result_type:
            case LookupResultType.BARE:
                assert source_result.collection_item is not None
                self.append_to_log_file(
                    "copy-collection-items.log",
                    [
                        f"Copying to {target_collection.collection}:"
                        f" {source_result.collection_item}"
                    ],
                )
                target_collection.collection.manager.add_bare_data(
                    category=BareDataCategory(
                        source_result.collection_item.category
                    ),
                    user=user,
                    workflow=workflow,
                    # To match update-collections-with-data, pass variables
                    # even if they were used to compute the item name.
                    data=expanded_variables,
                    name=item_name,
                    replace=replace,
                )
            case LookupResultType.ARTIFACT:
                assert source_result.artifact is not None
                self.append_to_log_file(
                    "copy-collection-items.log",
                    [
                        f"Copying to {target_collection.collection}:"
                        f" {source_result.artifact}"
                    ],
                )
                new_artifact = self.copy_artifact(
                    source_result.artifact, target_workspace
                )
                target_collection.collection.manager.add_artifact(
                    new_artifact,
                    user=user,
                    workflow=workflow,
                    variables=item_variables,
                    name=item_name,
                    replace=replace,
                )
            case _ as unreachable:
                raise AssertionError(
                    f"Unexpected lookup result type: {unreachable}"
                )

    @override
    def _execute(self) -> WorkRequest.Results:
        """Execute the task."""
        for copies in self.copies:
            for source_item in copies.source_items:
                try:
                    self.copy_item(
                        source_item,
                        copies.target_collection,
                        unembargo=copies.unembargo,
                        replace=copies.replace,
                        name_template=copies.name_template,
                        variables=copies.variables,
                        user=self.work_request.created_by,
                        workflow=self.work_request.parent,
                    )
                except Exception as e:
                    self.append_to_log_file(
                        "copy-collection-items.log", [f"Copy failed: {e}"]
                    )
                    raise

        return WorkRequest.Results.SUCCESS

    @override
    def get_label(self) -> str:
        """Return the task label."""
        target_collections = ", ".join(
            str(copies.target_collection.lookup) for copies in self.copies
        )
        return f"copy to {target_collections}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the collection UI helper."""

from typing import ClassVar, override

from django.test import RequestFactory, override_settings
from django.urls import reverse

from debusine.artifacts.models import CollectionCategory
from debusine.db.models import Collection, CollectionItem
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.places import IndexPlace, ResourcePlace
from debusine.web.views.ui.collections import CollectionItemUI, CollectionUI


class CollectionUITests(TestCase):
    """Test the :py:class:`CollectionUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    sid: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )

    def get_helper(self, collection: Collection) -> CollectionUI:
        """Instantiate a UI helper for a workspace."""
        factory = RequestFactory()
        request = factory.get("/")
        return CollectionUI(request, collection)

    def test_place(self) -> None:
        place = self.get_helper(self.sid).place
        self.assertIsInstance(place, ResourcePlace)

    def test_place_search(self) -> None:
        helper = self.get_helper(self.sid)
        place = helper.place_search
        self.assertIsInstance(place, IndexPlace)
        self.assertIs(
            place.parent,
            self.scenario.workspace.ui(helper.request).place_collection_list,
        )

    def test_place_update(self) -> None:
        helper = self.get_helper(self.sid)
        place = helper.place_update
        self.assertIsInstance(place, IndexPlace)
        self.assertIs(place.parent, helper.place)

    def test_place_delete(self) -> None:
        helper = self.get_helper(self.sid)
        place = helper.place_delete
        self.assertIsInstance(place, IndexPlace)
        self.assertIs(place.parent, helper.place)

    @override_settings(ROOT_URLCONF="debusine.web.archives.urls")
    def test_place_suite_root(self) -> None:
        helper = self.get_helper(self.sid)
        place = helper.place_suite_root
        self.assertEqual(place.title, f"Suite <code>{self.sid}</code>")
        self.assertEqual(
            place.url,
            reverse(
                "suite-root",
                kwargs={
                    "scope": self.sid.workspace.scope.name,
                    "workspace": self.sid.workspace.name,
                    "suite": self.sid.name,
                },
            ),
        )
        self.assertIs(
            place.parent,
            self.scenario.workspace.ui(helper.request).place_suite_list,
        )


class CollectionItemUITests(TestCase):
    """Test the :py:class:`CollectionItemUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    sid: ClassVar[Collection]
    item: ClassVar[CollectionItem]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        source = cls.playground.create_source_artifact()
        cls.sid = cls.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        cls.item = cls.sid.manager.add_artifact(
            source,
            user=cls.scenario.user,
            variables={"component": "main", "section": "extra"},
        )

    def get_helper(self, item: CollectionItem) -> CollectionItemUI:
        """Instantiate a UI helper for a workspace."""
        factory = RequestFactory()
        request = factory.get("/")
        return CollectionItemUI(request, item)

    def test_place(self) -> None:
        helper = self.get_helper(self.item)
        place = helper.place
        self.assertIsInstance(place, ResourcePlace)
        self.assertIs(place.parent, self.sid.ui(helper.request).place)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by DB-side debusine tasks."""

from collections.abc import Sequence
from typing import Self

import pydantic

from debusine.client.models import LookupResultType
from debusine.db.models import Artifact, Collection, CollectionItem
from debusine.db.models.lookups import LookupResult
from debusine.tasks.models import (
    BaseLabeledTaskInput,
    BaseTaskInputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupDict,
    LookupSingle,
)


class DBBaseTaskInputArtifact(BaseTaskInputArtifact):
    """BaseTaskInputArtifact caching a DB artifact."""

    #: Cached Artifact
    _artifact: Artifact | None = None

    @property
    def artifact(self) -> Artifact:
        """Return the cached artifact, or look it up by ID."""
        if self.artifact_id is None:
            raise AssertionError(
                "cannot access the Artifact of a task input without an ID"
            )
        if self._artifact is None:
            self._artifact = Artifact.objects.get(pk=self.artifact_id)
        return self._artifact

    @classmethod
    def from_artifact(cls, artifact: Artifact) -> Self:
        """Instantiate from an Artifact."""
        result = cls(
            artifact_id=artifact.id,
            category=artifact.category,
            data=artifact.create_data(),
        )
        result._artifact = artifact
        return result


class DBArtifactInfo(InputArtifactSingle):
    """Information about an artifact."""

    #: Cached Artifact
    _artifact: Artifact | None = None

    @property
    def artifact(self) -> Artifact:
        """Return the cached artifact, or look it up by ID."""
        if self.artifact_id is None:
            raise AssertionError(
                "cannot access the Artifact of a task input without an ID"
            )
        if self._artifact is None:
            self._artifact = Artifact.objects.get(pk=self.artifact_id)
        return self._artifact

    @classmethod
    def from_artifact(
        cls, artifact: Artifact, *, lookup: LookupSingle, label: str
    ) -> Self:
        """Instantiate from an Artifact."""
        result = cls(
            lookup=lookup,
            label=label,
            artifact_id=artifact.id,
            category=artifact.category,
            data=artifact.create_data(),
        )
        result._artifact = artifact
        return result


class DBInputArtifactMultiple(InputArtifactMultiple):
    """Represent declared multiple input artifacts for a task."""

    artifacts: Sequence[DBBaseTaskInputArtifact] = pydantic.Field(
        default_factory=list
    )


class DBInputCollectionSingle(InputCollectionSingle):
    """Information about a collection."""

    #: Cached Collection
    _collection: Collection | None = None

    @property
    def collection(self) -> Collection:
        """Return the cached collection, or look it up by ID."""
        if self._collection is None:
            self._collection = Collection.objects.get(pk=self.collection_id)
        return self._collection

    @classmethod
    def from_collection(
        cls, collection: Collection, *, lookup: LookupSingle, label: str
    ) -> Self:
        """Instantiate from a Collection."""
        result = cls(
            lookup=lookup,
            label=label,
            collection_id=collection.id,
            scope_name=collection.workspace.scope.name,
            workspace_name=collection.workspace.name,
            category=collection.category,
            name=collection.name,
            data=collection.data,
        )
        result._collection = collection
        return result


class DBInputLookupResult(BaseLabeledTaskInput):
    """
    Information about a generic lookup result.

    This may be a bare data item, an artifact, or a collection.
    """

    result_type: LookupResultType
    lookup: LookupSingle | LookupDict
    parent_collection_lookup: LookupSingle | None = None
    collection_item_id: int | None = None
    artifact_id: int | None = None
    collection_id: int | None = None

    _result: LookupResult | None = None

    @property
    def result(self) -> LookupResult:
        """Return the cached lookup result, or reconstruct it."""
        if self._result is None:
            self._result = LookupResult.create(
                result_type=self.result_type,
                lookup=self.lookup,
                parent_collection_lookup=self.parent_collection_lookup,
                collection_item=(
                    None
                    if self.collection_item_id is None
                    else CollectionItem.objects.get(pk=self.collection_item_id)
                ),
                artifact=(
                    None
                    if self.artifact_id is None
                    else Artifact.objects.get(pk=self.artifact_id)
                ),
                collection=(
                    None
                    if self.collection_id is None
                    else Collection.objects.get(pk=self.collection_id)
                ),
            )
        return self._result

    @classmethod
    def from_lookup_result(cls, result: LookupResult, *, label: str) -> Self:
        """Instantiate from a LookupResult."""
        instance = cls(
            label=label,
            result_type=result.result_type,
            lookup=result.lookup,
            parent_collection_lookup=result.parent_collection_lookup,
            collection_item_id=(
                None
                if result.collection_item is None
                else result.collection_item.id
            ),
            artifact_id=None if result.artifact is None else result.artifact.id,
            collection_id=(
                None if result.collection is None else result.collection.id
            ),
        )
        instance._result = result
        return instance



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Template tag library for archive pages."""

from typing import override

from django import template
from django.template import Context, TemplateSyntaxError
from django.template.base import Parser, Token, kwarg_re
from django.template.defaulttags import URLNode

from debusine.db.context import context
from debusine.server.scopes import urlconf_scope
from debusine.web.urls.url_utils import make_absolute_webui

register = template.Library()


class DebusineURLNode(URLNode):
    """Extend URLNode to generate URLs to the main Debusine web UI."""

    @override
    def render(self, ctx: Context) -> str:
        with urlconf_scope(context.require_scope().name):
            path = super().render(ctx)
            if self.asvar:
                ctx[self.asvar] = make_absolute_webui(ctx[self.asvar])
                return path
            else:
                return make_absolute_webui(path)


@register.tag
def debusine_url(parser: Parser, token: Token) -> DebusineURLNode:
    """Behave like the ``{% url %}`` tag, but link to the main web UI."""
    # This is copied from Django's template.defaulttags.url, except for the
    # node class used at the end. It needs to be kept in sync if new versions
    # of Django change the behaviour or the implementation of {% url %}
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError(
            f"'{bits[0]}' takes at least one argument, a URL pattern name."
        )
    viewname = parser.compile_filter(bits[1])
    args = []
    kwargs = {}
    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == "as":
        asvar = bits[-1]
        bits = bits[:-2]

    for bit in bits:
        match = kwarg_re.match(bit)
        if not match:
            raise TemplateSyntaxError("Malformed arguments to url tag")
        name, value = match.groups()
        if name:
            kwargs[name] = parser.compile_filter(value)
        else:
            args.append(parser.compile_filter(value))

    return DebusineURLNode(viewname, args, kwargs, asvar)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the enroll views."""

import asyncio
import contextlib
from collections.abc import AsyncIterator, Generator
from typing import Any, ClassVar, Never, override
from unittest import mock

from asgiref.sync import sync_to_async
from channels.layers import get_channel_layer
from django.db import transaction
from django.http import StreamingHttpResponse
from django.test import AsyncClient, Client
from django.urls import reverse
from rest_framework import status

from debusine.client.models import (
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
)
from debusine.db.models.auth import ClientEnroll
from debusine.test.django import TestCase, TestResponseType


class EnrollViewTests(TestCase):
    """Tests for :py:class:`EnrollView`."""

    payload: ClassVar[EnrollPayload]
    url: ClassVar[str]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.payload = EnrollPayload(
            nonce="12345678",
            challenge="correct horse battery staple",
            hostname="hostname",
            scope="scope",
        )
        cls.url = reverse("api:enroll")

    @contextlib.contextmanager
    def assertValidationSuccessful(self) -> Generator[None]:
        """Ensure EnrollView passes validation."""

        class Reached(BaseException):
            """Thrown when the target code is reached."""

        with (
            mock.patch(
                "debusine.server.views.enroll.EnrollView.long_poll",
                side_effect=Reached,
            ),
            self.assertRaises(Reached),
        ):
            yield

    async def assertResponsePayload(
        self, response: StreamingHttpResponse
    ) -> EnrollConfirmPayload:
        """Consume streaming content and return the parsed payload."""
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        # Django's typing currently do not seem to contemplate async streaming
        # content
        assert isinstance(response.streaming_content, AsyncIterator)
        chunks = [x async for x in response.streaming_content]
        return EnrollConfirmPayload.model_validate_json(b"".join(chunks))

    def _post(
        self,
        payload: EnrollPayload | dict[str, Any] | str | None,
        content_type: str = "application/json",
    ) -> "TestResponseType | StreamingHttpResponse":
        match payload:
            case EnrollPayload():
                return self.client.post(
                    self.url,
                    data=payload.model_dump(),
                    content_type=content_type,
                )
            case dict():
                return self.client.post(
                    self.url, data=payload, content_type=content_type
                )
            case "":
                # Django's test client does 'if data' instead of 'if data is
                # None', so we need to special-case sending an empty payload
                return self.client.post(
                    self.url,
                    data=b"",
                    content_type=content_type,
                    headers={
                        "Content-Length": "0",
                        "Content-Type": content_type,
                    },
                )
            case _:
                return self.client.post(
                    self.url, data=payload, content_type=content_type
                )

    async def _apost(self, payload: EnrollPayload) -> StreamingHttpResponse:
        """
        POST the payload, and wait until the view waits for confirmation.

        :returns: the response task, paused at ``get_channel_layer``.
        """
        client = AsyncClient()
        res = await client.post(
            self.url,
            data=payload.model_dump(),
            content_type="application/json",
        )
        assert isinstance(res, StreamingHttpResponse)
        return res

    def test_invalid_methods(self) -> None:
        """Only POST is accepted."""
        for method in "get", "put", "patch", "delete", "head":
            with self.subTest(method=method):
                response = getattr(self.client, method)(self.url)
                self.assertEqual(
                    response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED
                )

    def test_unauthenticated(self) -> None:
        """Authentication is not needed."""
        with self.assertValidationSuccessful():
            self._post(self.payload)

    def test_csrf_exempt(self) -> None:
        """CSRF token is not needed."""
        client = Client(enforce_csrf_checks=True)
        with self.assertValidationSuccessful():
            client.post(
                self.url,
                data=self.payload.model_dump(),
                content_type="application/json",
            )

    def test_request_content_type(self) -> None:
        """Only application/json requests are accepted."""
        for content_type in (
            "text/json",
            "text/plain",
            "application/x-www-form-urlencoded",
            "multipart/form-data",
        ):
            with self.subTest(content_type=content_type):
                response = self._post(self.payload, content_type=content_type)
                self.assertResponseProblem(
                    response,
                    "only JSON data is accepted",
                )

    def test_request_body_length_cap(self) -> None:
        """Request body length is capped."""
        response = self._post("*" * 4097)
        self.assertResponseProblem(response, "JSON payload too big")

    def test_request_body_valid_json(self) -> None:
        """Request body must be valid JSON."""
        for body in "{", "", "[", "foo":
            with self.subTest(body=body):
                response = self._post(body)
                self.assertResponseProblem(
                    response, "payload is not valid JSON"
                )

    def test_request_body_valid_payload(self) -> None:
        """Request body must be a valid EnrollPayload."""
        for body in (
            {},
            "[]",
            "null",
            "42",
            {"nonce": "foobarbaz"},
            {"nonce": "foo", "challenge": "too short"},
        ):
            with self.subTest(body=body):
                response = self._post(body)
                self.assertResponseProblem(response, "payload data is invalid")

    async def test_payload_in_database(self) -> None:
        """Payload is stored in the database."""
        response = await self._apost(self.payload)

        # Payload is in database while the post is waiting for confirmation
        ce = await ClientEnroll.objects.aget(nonce=self.payload.nonce)
        self.assertEqual(ce.nonce, self.payload.nonce)
        self.assertEqual(ce.payload, self.payload.model_dump())

        # Send confirmation
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        confirmation = EnrollConfirmPayload(
            outcome=EnrollOutcome.CONFIRM, token="a" * 32
        )
        await channel_layer.send(
            f"enroll.{self.payload.nonce}", confirmation.model_dump()
        )

        # The payload is in the database while the response is streaming (that
        # is, waiting/long polling)
        await ClientEnroll.objects.aget(nonce=self.payload.nonce)

        # Consume the streaming output
        payload = await self.assertResponsePayload(response)

        # The payload has been deleted from the database
        with self.assertRaises(ClientEnroll.DoesNotExist):
            await ClientEnroll.objects.aget(nonce=self.payload.nonce)

        self.assertEqual(payload.token, "a" * 32)

    async def test_duplicate_request(self) -> None:
        """Payload is stored in the database."""
        # The second request is isolated in a separate executor, to avoid a
        # query error that invalidates the transaction of the first one

        def duplicate_request() -> None:
            with transaction.atomic():
                self.assertResponseProblem(
                    self._post(self.payload),
                    "duplicate or invalid request received",
                )

        response = await self._apost(self.payload)

        # The second response fails
        await sync_to_async(duplicate_request)()

        # Send confirmation to complete the first response
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        confirmation = EnrollConfirmPayload(
            outcome=EnrollOutcome.CONFIRM, token="a" * 32
        )
        await channel_layer.send(
            f"enroll.{self.payload.nonce}", confirmation.model_dump()
        )
        payload = await self.assertResponsePayload(response)
        self.assertEqual(payload.token, "a" * 32)

    async def test_client_disconnect(self) -> None:
        """Client disconnects during streaming."""
        # Replace the channel receive result with a future that we can cancel,
        # to simulate a client disconnect

        class _MockChannelLayer:
            def __init__(self) -> None:
                self.future: asyncio.Future[None] = asyncio.Future()

            async def receive(
                self,
                channel_name: str,  # noqa: ARG002, U100
            ) -> Never:
                await self.future
                raise AssertionError(
                    "the future should have raised CancelledError"
                )

        mock_channel_layer = _MockChannelLayer()

        with mock.patch(
            "debusine.server.views.enroll.get_channel_layer",
            return_value=mock_channel_layer,
        ):
            response = await self._apost(self.payload)

            # Payload is in the database
            await ClientEnroll.objects.aget(nonce=self.payload.nonce)

            # Simulate a browser disconnect
            mock_channel_layer.future.cancel()

            # Consume streaming content as a replacement for the ASGI server
            # doing its thing
            # TODO: is there a better way to simulate a client disconnect?
            with self.assertRaises(asyncio.CancelledError):
                assert isinstance(response.streaming_content, AsyncIterator)
                async for _chunk in response.streaming_content:
                    pass  # pragma: no cover

            # The payload has been deleted from the database
            with self.assertRaises(ClientEnroll.DoesNotExist):
                await ClientEnroll.objects.aget(nonce=self.payload.nonce)

    async def test_keepalive(self) -> None:
        """Test data sent to keep the connection alive during the long poll."""
        # Set interval timeout to "immediate"
        with mock.patch(
            "debusine.server.views.enroll.EnrollView.PING_INTERVAL", 0
        ):
            response = await self._apost(self.payload)

            # Before confirmation arrives, the line is kept alive with newlines
            assert isinstance(response.streaming_content, AsyncIterator)
            chunk = await anext(response.streaming_content)
            self.assertEqual(chunk, b"\n")

            # Send confirmation
            channel_layer = get_channel_layer()
            assert channel_layer is not None
            confirmation = EnrollConfirmPayload(
                outcome=EnrollOutcome.CONFIRM, token="a" * 32
            )

            await channel_layer.send(
                f"enroll.{self.payload.nonce}", confirmation.model_dump()
            )

            payload = await self.assertResponsePayload(response)
            self.assertEqual(payload.token, "a" * 32)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine dput-ng uploader implementation."""

from pathlib import Path
from typing import Any
from urllib.parse import quote, urljoin

from dput.core import logger
from dput.exceptions import UploadException
from dput.uploader import AbstractUploader

from debusine.client.client_utils import prepare_changes_for_upload
from debusine.client.dput_ng.dput_ng_utils import make_debusine_client
from debusine.client.models import RelationType


class DebusineUploader(AbstractUploader):  # type: ignore[misc]
    """An interface to upload files to debusine."""

    def initialize(self, **kwargs: Any) -> None:
        """See :py:meth:`AbstractUploader.initialize`."""
        self.scope = self._config["debusine_scope"]
        self.workspace = self._config["debusine_workspace"]
        self.debusine = make_debusine_client(self._config)
        self.paths: list[Path] = []

    def upload_file(
        self, filename: str, upload_filename: str | None = None
    ) -> None:
        """See :py:meth:`AbstractUploader.upload_file`."""
        if upload_filename is not None:
            raise NotImplementedError(
                "Debusine does not support commands files."
            )

        path = Path(filename)
        for old_path in self.paths:
            if old_path.suffix == path.suffix and path.suffix == ".changes":
                raise RuntimeError("Cannot upload multiple .changes files")

        self.paths.append(path)

        if path.suffix == ".changes":
            # TODO: We should check that the files expected by these
            # artifacts match the set of files referenced by self.paths.
            artifacts = prepare_changes_for_upload(path)
            try:
                uploaded = [
                    self.debusine.upload_artifact(
                        artifact, workspace=self.workspace
                    )
                    for artifact in artifacts
                ]
                primary_artifact = uploaded[0]
                for related_artifact in uploaded[1:]:
                    for relation_type in (
                        RelationType.EXTENDS,
                        RelationType.RELATES_TO,
                    ):
                        self.debusine.relation_create(
                            primary_artifact.id,
                            related_artifact.id,
                            relation_type,
                        )
            except Exception as e:
                raise UploadException(str(e)) from e
            logger.info(
                "Created artifact: %s",
                # TODO: Ideally we wouldn't have to hardcode the URL
                # structure here.
                urljoin(
                    self.debusine.base_api_url,
                    f"/{quote(self.scope)}/{quote(self.workspace)}/artifact/"
                    f"{primary_artifact.id}/",
                ),
            )

            # Save artifact IDs in the in-memory copy of the profile, for
            # use by hooks.
            artifact_ids = {"changes": primary_artifact.id}
            self._config["debusine_artifact_ids"] = artifact_ids



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Django Application Configuration for the server application."""

from typing import override

from django.apps import AppConfig
from django.conf import settings
from prometheus_client import values

from debusine.server.prometheus_client.values import RedisValue


class ServerConfig(AppConfig):
    """Django's AppConfig for the server application."""

    name = 'debusine.server'

    @override
    def ready(self) -> None:
        """Mark workers as disconnected. Runs on Django startup."""
        # Import the provisioner and scheduler so that their Celery tasks
        # and signal receivers are registered.
        import debusine.server.provisioning  # noqa: F401
        import debusine.server.scheduler  # noqa: F401

        # Import various tasks so that they are registered.
        import debusine.server.tasks  # noqa: F401
        import debusine.server.tasks.wait  # noqa: F401
        import debusine.signing.tasks  # noqa: F401


if settings.PROMETHEUS_REDIS_URL:  # pragma: no cover
    # Hook our Redis prometheus_client Values implementation in before any
    # metrics are created by importing a module
    values.ValueClass = RedisValue()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for Celery integration."""

import logging
from typing import Any, override
from unittest import mock

from django.db import connections, transaction

from debusine.db.models import WorkRequest, Worker
from debusine.server.celery import WorkRequestNotPending, run_server_task
from debusine.server.tasks.models import ServerNoopData
from debusine.server.tasks.tests.helpers import SampleBaseServerTask
from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskData,
    OutputData,
    OutputDataError,
)
from debusine.test.django import TestCase, TransactionTestCase


class SampleTaskNotManagingTransactions(
    SampleBaseServerTask[BaseTaskData, BaseDynamicTaskData]
):
    """A sample server task that does not manage its own transactions."""

    TASK_VERSION = 1

    @override
    def _execute(self) -> WorkRequest.Results:
        return (
            WorkRequest.Results.SUCCESS
            if connections["default"].in_atomic_block
            else WorkRequest.Results.FAILURE
        )


class SampleTaskManagingTransactions(
    SampleBaseServerTask[BaseTaskData, BaseDynamicTaskData]
):
    """A sample server task that manages its own transactions."""

    TASK_VERSION = 1
    TASK_MANAGES_TRANSACTIONS = True

    @override
    def _execute(self) -> WorkRequest.Results:
        return (
            WorkRequest.Results.SUCCESS
            if not connections["default"].in_atomic_block
            else WorkRequest.Results.FAILURE
        )


class RunServerTaskTests(TestCase):
    """Unit tests for :py:func:`run_server_task`."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.worker = Worker.objects.get_or_create_celery()

    def create_assigned_work_request(
        self, task_name: str = "servernoop", **kwargs: Any
    ) -> WorkRequest:
        """Create a work request and assign it to self.worker."""
        work_request = self.playground.create_server_task(
            task_name=task_name, mark_pending=True, worker=self.worker, **kwargs
        )
        return work_request

    def test_no_work_request(self) -> None:
        """The Celery task fails if the work request does not exist."""
        work_request_id = self.playground.create_worker_task().id + 1
        with self.assertLogsContains(
            f"Work request {work_request_id} does not exist",
            logger="debusine.server.celery",
            level=logging.ERROR,
        ):
            result = run_server_task.apply(args=(work_request_id,))
        with self.assertRaises(WorkRequest.DoesNotExist):
            result.get()

    def test_not_pending(self) -> None:
        """The Celery task fails if the work request is not pending."""
        work_request = self.create_assigned_work_request(
            result=WorkRequest.Results.ERROR
        )
        with self.assertLogsContains(
            f"Work request {work_request.id} is in status completed, not "
            f"pending",
            logger="debusine.server.celery",
            level=logging.ERROR,
        ):
            result = run_server_task.apply(args=(work_request.id,))
        with self.assertRaises(WorkRequestNotPending):
            result.get()

    def test_bad_task_name(self) -> None:
        """The Celery task logs but does not fail if given a bad task name."""
        work_request = self.create_assigned_work_request()
        work_request.task_name = "nonexistent"
        work_request.save()
        expected_message = "'nonexistent' is not a registered Server task_name"
        with (
            self.assertLogsContains(
                f"Error running work request Server/nonexistent "
                f"({work_request.id}): {expected_message}",
                logger="debusine.server.celery",
                level=logging.ERROR,
            ),
            transaction.atomic(),
        ):
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="setup-failed"
                    )
                ]
            ),
        )

    def test_bad_task_data(self) -> None:
        """The Celery task logs but does not fail if given bad task data."""
        work_request = self.create_assigned_work_request()
        work_request.task_data = {"nonexistent": True}
        work_request.configured_task_data = {"nonexistent": True}
        work_request.save()
        expected_message = (
            "Failed to configure: 1 validation error for ServerNoopData"
        )
        with (
            self.assertLogsContains(
                f"Error running work request Server/servernoop "
                f"({work_request.id}): {expected_message}",
                logger="debusine.server.celery",
                level=logging.ERROR,
            ),
            transaction.atomic(),
        ):
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        assert work_request.output_data is not None
        self.assertIsInstance(work_request.output_data, OutputData)
        assert work_request.output_data.errors is not None
        self.assertEqual(len(work_request.output_data.errors), 1)
        self.assertTrue(
            work_request.output_data.errors[0].message.startswith(
                expected_message
            )
        )
        self.assertEqual(
            work_request.output_data.errors[0].code, "configure-failed"
        )

    def test_external_task(self) -> None:
        """The Celery task logs but does not fail on an external task."""
        work_request = self.playground.create_worker_task(
            task_name="noop",
            task_data={"result": WorkRequest.Results.SUCCESS},
            mark_pending=True,
            worker=self.worker,
        )
        expected_message = "Cannot run on a Celery worker"
        with (
            self.assertLogsContains(
                f"Error running work request Worker/noop ({work_request.id}): "
                f"{expected_message}",
                logger="debusine.server.celery",
                level=logging.ERROR,
            ),
            transaction.atomic(),
        ):
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="wrong-task-type"
                    )
                ]
            ),
        )

    def test_task_raises_exception(self) -> None:
        """The Celery task logs but does not fail if the task raises."""
        work_request = self.create_assigned_work_request(
            task_data=ServerNoopData(exception=True),
        )
        expected_message = "Execution failed: Client requested an exception"
        with (
            self.assertLogsContains(
                f"Error running work request Server/servernoop "
                f"({work_request.id}): {expected_message}",
                logger="debusine.server.celery",
                level=logging.ERROR,
            ),
            transaction.atomic(),
        ):
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            work_request.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="execute-failed"
                    )
                ]
            ),
        )

    def test_task_aborted(self) -> None:
        """If the task is aborted, the Celery task logs but does not fail."""
        work_request = self.create_assigned_work_request()
        with (
            self.assertLogsContains(
                "Task: servernoop has been aborted",
                logger="debusine.server.celery",
                level=logging.INFO,
            ),
            mock.patch(
                "debusine.server.tasks.noop.ServerNoop.execute",
                lambda task: task.abort(),
            ),
            transaction.atomic(),
        ):
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.FAILURE)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(work_request.result, WorkRequest.Results.NONE)
        self.assertIsNone(work_request.output_data)

    def test_task_returns_failure(self) -> None:
        """If the task returns FAILURE, the Celery task records that."""
        work_request = self.create_assigned_work_request(
            task_data=ServerNoopData(result=WorkRequest.Results.FAILURE),
        )
        with transaction.atomic():
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.FAILURE)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.FAILURE)
        self.assertIsNone(work_request.output_data)

    def test_task_returns_true(self) -> None:
        """If the task returns True, the Celery task records that."""
        work_request = self.create_assigned_work_request(
            task_data=ServerNoopData(result=WorkRequest.Results.SUCCESS),
        )
        with transaction.atomic():
            result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.SUCCESS)
        work_request.refresh_from_db()
        self.assertIsNotNone(work_request.started_at)
        self.assertIsNotNone(work_request.completed_at)
        self.assertEqual(work_request.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(work_request.result, WorkRequest.Results.SUCCESS)
        self.assertIsNone(work_request.output_data)


class RunServerTaskTransactionTests(TransactionTestCase):
    """Unit tests for :py:func:`run_server_task`."""

    def test_task_not_managing_transactions(self) -> None:
        """An atomic block is used for tasks that do not manage transactions."""
        worker = Worker.objects.get_or_create_celery()
        work_request = self.playground.create_server_task(
            task_name="sampletasknotmanagingtransactions",
            task_data={},
            mark_pending=True,
            worker=worker,
        )
        result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.SUCCESS)

    def test_task_managing_transactions(self) -> None:
        """An atomic block is not used for tasks that manage transactions."""
        worker = Worker.objects.get_or_create_celery()
        work_request = self.playground.create_server_task(
            task_name="sampletaskmanagingtransactions",
            task_data={},
            mark_pending=True,
            worker=worker,
        )
        result = run_server_task.apply(args=(work_request.id,))
        self.assertEqual(result.get(), WorkRequest.Results.SUCCESS)

    def test_locks_workflow_before_completion(self) -> None:
        workflow = self.playground.create_workflow()
        server_noop = workflow.create_child_server(
            task_name="servernoop",
            task_data=ServerNoopData(result=WorkRequest.Results.SUCCESS),
        )
        noop = workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(
            server_noop, mark_pending=True, assign_new_worker=True
        )
        unrelated = self.playground.create_worker_task()

        with mock.patch.object(
            WorkRequest,
            "mark_completed",
            side_effect=(
                lambda *args, **kwargs: self.assert_work_requests_locked(
                    locked=[workflow, server_noop, noop], unlocked=[unrelated]
                )
            ),
        ):
            result = run_server_task.apply(args=(server_noop.id,))

        self.assertEqual(result.get(), WorkRequest.Results.SUCCESS)

    def test_locks_workflow_before_completion_error(self) -> None:
        workflow = self.playground.create_workflow()
        server_noop = workflow.create_child_server(
            task_name="servernoop", task_data=ServerNoopData(exception=True)
        )
        noop = workflow.create_child_worker(task_name="noop")
        self.playground.advance_work_request(
            server_noop, mark_pending=True, assign_new_worker=True
        )
        unrelated = self.playground.create_worker_task()

        with mock.patch.object(
            WorkRequest,
            "mark_completed",
            side_effect=(
                lambda *args, **kwargs: self.assert_work_requests_locked(
                    locked=[workflow, server_noop, noop], unlocked=[unrelated]
                )
            ),
        ):
            result = run_server_task.apply(args=(server_noop.id,))

        self.assertEqual(result.get(), WorkRequest.Results.ERROR)
